# SPDX-License-Identifier: MIT
"""Point score, madison and omnium handler for trackmeet."""

import gtk
import glib
import gobject
import os
import logging
import csv

import metarace
from metarace import timy
from metarace import scbwin
from metarace import tod
from metarace import unt4
from metarace import uiutil
from metarace import riderdb
from metarace import strops
from metarace import report
from metarace import jsonconfig

_log = logging.getLogger(u'metarace.ps')
_log.setLevel(logging.DEBUG)

# config version string
EVENT_ID = u'ps-2.0'

# Model columns
SPRINT_COL_ID = 0
SPRINT_COL_LABEL = 1
SPRINT_COL_200 = 2
SPRINT_COL_SPLIT = 3
SPRINT_COL_PLACES = 4
SPRINT_COL_POINTS = 5

RES_COL_BIB = 0
RES_COL_FIRST = 1
RES_COL_LAST = 2
RES_COL_CLUB = 3
RES_COL_INRACE = 4
RES_COL_POINTS = 5
RES_COL_LAPS = 6
RES_COL_TOTAL = 7
RES_COL_PLACE = 8
RES_COL_FINAL = 9
RES_COL_INFO = 10
RES_COL_STPTS = 11

# scb consts

SPRINT_PLACE_DELAY = 3  # 3 seconds per place
SPRINT_PLACE_DELAY_MAX = 11  # to a maximum of 11

# scb function key mappings
key_startlist = u'F3'
key_results = u'F4'

# timing function key mappings
key_armstart = u'F5'
key_showtimer = u'F6'
key_armfinish = u'F9'
key_lapdown = u'F11'

# extended function key mappings
key_abort = u'F5'
key_falsestart = u'F6'


class ps(object):
    """Data handling for point score omnium and Madison races."""

    def loadconfig(self):
        """Load race config from disk."""
        self.riders.clear()
        self.sprints.clear()
        self.sprintpoints = {}
        definomnium = False
        defsprintlaps = u''
        defscoretype = u'points'
        defmasterslaps = u'No'
        if self.evtype == u'madison':
            defscoretype = u'madison'
            defmasterslaps = u'No'
        elif self.evtype == u'omnium':
            definomnium = True
            defsprintlaps = u'scr tmp elm'
            self.laplabels = {
                u'scr': u'Scratch',
                u'tmp': u'Tempo',
                u'elm': u'Elimination'
            }
            self.sprintpoints = {
                u'scr':
                u'40 38 36 34 32 30 28 26 24 22 20 18 16 14 12 10 8 6 4 2 1 1 1 1 1 1',
                u'tmp':
                u'40 38 36 34 32 30 28 26 24 22 20 18 16 14 12 10 8 6 4 2 1 1 1 1 1 1',
                u'elm':
                u'40 38 36 34 32 30 28 26 24 22 20 18 16 14 12 10 8 6 4 2 1 1 1 1 1 1',
            }

        cr = jsonconfig.config({
            u'event': {
                u'startlist': u'',
                u'id': EVENT_ID,
                u'start': None,
                u'lstart': None,
                u'finish': None,
                u'comments': [],
                u'sprintlaps': defsprintlaps,
                u'distance': u'',
                u'runlap': u'',
                u'distunits': u'laps',
                u'masterslaps': defmasterslaps,
                u'inomnium': definomnium,
                u'showinfo': True,
                u'autospec': u'',
                u'scoring': defscoretype
            }
        })
        cr.add_section(u'event')
        cr.add_section(u'sprintplaces')
        cr.add_section(u'sprintpoints')
        cr.add_section(u'sprintsource')
        cr.add_section(u'laplabels')
        cr.add_section(u'points')
        if os.path.exists(self.configfile):
            try:
                with open(self.configfile, 'rb') as f:
                    cr.read(f)
            except Exception as e:
                _log.error(u'Unable to read config: %s', e)
        else:
            _log.info(u'%r not found, loading defaults', self.configfile)
        self.inomnium = cr.get_bool(u'event', u'inomnium')
        if self.inomnium:
            self.seedsrc = 1  # fetch start list seeding from omnium

        for r in cr.get(u'event', u'startlist').split():
            nr = [r, u'', u'', u'', True, 0, 0, 0, u'', -1, u'', 0]
            if cr.has_option(u'points', r):
                ril = cr.get(u'points', r)
                if len(ril) >= 1:
                    nr[RES_COL_INRACE] = strops.confopt_bool(ril[0])
                if len(ril) >= 3:
                    try:
                        nr[RES_COL_LAPS] = int(ril[2])
                    except ValueError:
                        pass
                if len(ril) >= 4:
                    nr[RES_COL_INFO] = ril[3]
                if len(ril) >= 5:
                    spts = ril[4]
                    if spts.isdigit():
                        nr[RES_COL_STPTS] = int(spts)

            dbr = self.meet.rdb.getrider(r, self.series)
            if dbr is not None:
                for i in range(1, 4):
                    nr[i] = self.meet.rdb.getvalue(dbr, i)
            self.riders.append(nr)
        if cr.get(u'event', u'scoring').lower() == u'madison':
            self.scoring = u'madison'
        else:
            self.scoring = u'points'
        self.type_lbl.set_text(self.scoring.capitalize())

        # race infos
        self.comments = cr.get(u'event', u'comments')

        self.autospec = cr.get(u'event', u'autospec')
        self.distance = strops.confopt_dist(cr.get(u'event', u'distance'))
        self.units = strops.confopt_distunits(cr.get(u'event', u'distunits'))
        self.runlap = cr.get_posint(u'event', u'runlap')
        self.masterslaps = cr.get_bool(u'event', u'masterslaps')
        # override laps from event listings
        if not self.onestart and self.event[u'laps']:
            self.units = u'laps'
            self.distance = strops.confopt_posint(self.event[u'laps'],
                                                  self.distance)

        self.reset_lappoints()
        slt = cr.get(u'event', u'sprintlaps')
        self.sprintlaps = strops.reformat_biblist(slt)

        # load any special purpose sprint points
        for sid in cr.options(u'sprintpoints'):
            self.sprintpoints[sid] = cr.get(u'sprintpoints', sid)

        # load lap labels
        for sid in cr.options(u'laplabels'):
            self.laplabels[sid] = cr.get(u'laplabels', sid)

        # load any autospec'd sprint results
        for sid in cr.options(u'sprintsource'):
            self.sprintsource[sid] = cr.get(u'sprintsource', sid)

        self.sprint_model_init()

        oft = 0
        for s in self.sprints:
            places = u''
            sid = s[SPRINT_COL_ID].decode(u'utf-8')
            if cr.has_option(u'sprintplaces', sid):
                sp = cr.get(u'sprintplaces', sid)
                #_log.debug(u'sprintplaces = %r', sp)
                places = strops.reformat_placelist(sp)
                if len(places) > 0:
                    oft += 1
            s[SPRINT_COL_PLACES] = places
            if cr.has_option(u'sprintplaces', sid + u'_200'):
                s[SPRINT_COL_200] = tod.mktod(
                    cr.get(u'sprintplaces', sid + u'_200'))
            if cr.has_option(u'sprintplaces', sid + u'_split'):
                s[SPRINT_COL_SPLIT] = tod.mktod(
                    cr.get(u'sprintplaces', sid + u'_split'))
        if oft > 0:
            if oft >= len(self.sprints):
                oft = len(self.sprints) - 1
            self.ctrl_place_combo.set_active(oft)
            self.onestart = True

        ## for omnium - look up the places from event links if present
        if self.inomnium:
            for s in self.sprints:
                sid = s[SPRINT_COL_ID].decode(u'utf-8')
                if sid in self.sprintsource:
                    splac = self.meet.autoplace_riders(self,
                                                       self.sprintsource[sid],
                                                       final=True)
                    _log.debug(u'Loaded %r places from event %r: %r', sid,
                               self.sprintsource[sid], splac)
                    if splac:
                        s[SPRINT_COL_PLACES] = splac
        self.recalculate()

        self.info_expand.set_expanded(cr.get_bool(u'event', u'showinfo'))
        self.set_start(cr.get(u'event', u'start'), cr.get(u'event', u'lstart'))
        self.set_finish(cr.get(u'event', u'finish'))
        self.set_elapsed()

        # after load, add auto if required
        if not self.onestart and self.autospec:
            _log.debug(u'Fetching starters using autospec=%r with seedsrc=%r',
                       self.autospec, self.seedsrc)
            self.meet.autostart_riders(self, self.autospec, self.seedsrc)

        # After load complete - check config and report.
        eid = cr.get(u'event', u'id')
        if eid and eid != EVENT_ID:
            _log.info(u'Event config mismatch: %r != %r', eid, EVENT_ID)

    def get_startlist(self):
        """Return a list of bibs in the rider model."""
        ret = []
        for r in self.riders:
            ret.append(r[RES_COL_BIB].decode(u'utf-8'))
        return u' '.join(ret)

    def saveconfig(self):
        """Save race to disk."""
        if self.readonly:
            _log.error('Attempt to save readonly event')
            return
        cw = jsonconfig.config()
        cw.add_section(u'event')
        if self.start is not None:
            cw.set(u'event', u'start', self.start.rawtime())
        if self.lstart is not None:
            cw.set(u'event', u'lstart', self.lstart.rawtime())
        if self.finish is not None:
            cw.set(u'event', u'finish', self.finish.rawtime())
        cw.set(u'event', u'startlist', self.get_startlist())
        cw.set(u'event', u'showinfo', self.info_expand.get_expanded())
        cw.set(u'event', u'distance', self.distance)
        cw.set(u'event', u'distunits', self.units)
        cw.set(u'event', u'scoring', self.scoring)
        if self.runlap is not None:
            cw.set(u'event', u'runlap', self.runlap)
        cw.set(u'event', u'masterslaps', self.masterslaps)
        cw.set(u'event', u'autospec', self.autospec)
        cw.set(u'event', u'inomnium', self.inomnium)
        cw.set(u'event', u'sprintlaps', self.sprintlaps)
        cw.set(u'event', u'comments', self.comments)

        cw.add_section(u'sprintplaces')
        cw.add_section(u'sprintpoints')
        cw.add_section(u'sprintsource')
        cw.add_section(u'laplabels')
        for s in self.sprints:
            sid = s[SPRINT_COL_ID].decode(u'utf-8')
            cw.set('sprintplaces', sid, s[SPRINT_COL_PLACES].decode(u'utf-8'))
            if s[SPRINT_COL_200] is not None:
                cw.set(u'sprintplaces', sid + u'_200',
                       s[SPRINT_COL_200].rawtime())
            if s[SPRINT_COL_SPLIT] is not None:
                cw.set(u'sprintplaces', sid + u'_split',
                       s[SPRINT_COL_SPLIT].rawtime())
            if s[SPRINT_COL_POINTS] is not None:
                cw.set(u'sprintpoints', sid,
                       u' '.join(map(unicode, s[SPRINT_COL_POINTS])))
            if sid in self.laplabels:
                cw.set(u'laplabels', sid, self.laplabels[sid])
            if sid in self.sprintsource:
                cw.set(u'sprintsource', sid, self.sprintsource[sid])

        # rider result section
        cw.add_section(u'points')
        for r in self.riders:
            slice = [
                r[RES_COL_INRACE],
                unicode(r[RES_COL_POINTS]),
                unicode(r[RES_COL_LAPS]),
                unicode(r[RES_COL_INFO]),
                unicode(r[RES_COL_STPTS])
            ]
            cw.set(u'points', r[RES_COL_BIB], slice)

        cw.set(u'event', u'id', EVENT_ID)
        _log.debug(u'Saving points config %r', self.configfile)
        with metarace.savefile(self.configfile) as f:
            cw.write(f)

    def result_gen(self):
        """Generator function to export a final result."""
        fl = None
        ll = None
        for r in self.riders:
            bib = r[RES_COL_BIB]
            rank = None
            info = None
            if r[RES_COL_TOTAL] == 0:
                info = '-'
            else:
                info = str(r[RES_COL_TOTAL])
            if self.onestart:
                if r[RES_COL_INRACE]:
                    if r[RES_COL_PLACE] is not None and r[RES_COL_PLACE] != '':
                        rank = int(r[RES_COL_PLACE])
                else:
                    if r[RES_COL_INFO] in ['dns', 'dsq']:
                        rank = r[RES_COL_INFO]
                    else:
                        if r[RES_COL_INFO].isdigit():
                            rank = int(r[RES_COL_INFO])
                        else:
                            rank = 'dnf'  # ps only handle did not finish

            if self.scoring == 'madison':
                laps = r[RES_COL_LAPS]
                if fl is None:
                    fl = laps  # anddetermine laps down
                if ll is not None:
                    down = fl - laps
                    if ll != laps:
                        yield [
                            '', ' ',
                            str(down) + ' Lap' + strops.plural(down) +
                            ' Behind', ''
                        ]
                ll = laps

            time = None
            yield [bib, rank, time, info]

    def result_report(self, recurse=False):
        """Return a list of report sections containing the race result."""
        self.recalculate()
        ret = []
        sec = report.section()
        sec.heading = 'Event ' + self.evno + ': ' + ' '.join(
            [self.event[u'pref'], self.event[u'info']]).strip()
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        sec.units = 'Pts'
        fs = ''
        if self.finish is not None:
            fs = self.time_lbl.get_text().strip()
        fl = None
        ll = None
        for r in self.riders:
            rno = r[RES_COL_BIB].decode('utf-8')
            rh = self.meet.newgetrider(rno, self.series)
            rname = u''
            if rh is not None:
                rname = rh[u'namestr']
            plstr = ''
            rcat = None
            if self.event[u'cate']:
                if rname is not None and rh[u'cat']:
                    rcat = rh[u'cat']
            if self.inomnium:
                rcat = None
            if rh is not None and rh[u'ucicode']:
                rcat = rh[u'ucicode']  # overwrite by force
            if self.onestart and r[RES_COL_PLACE] is not None:
                plstr = r[RES_COL_PLACE]
                if r[RES_COL_PLACE].isdigit():
                    plstr += '.'
                elif r[RES_COL_INFO] in [u'dns', u'dsq']:
                    plstr = r[RES_COL_INFO]
                elif r[RES_COL_INFO] and r[RES_COL_INFO].isdigit():
                    plstr = r[RES_COL_INFO] + u'.'
                ptstr = ''
                if r[RES_COL_TOTAL] != 0 and r[RES_COL_INRACE]:
                    ptstr = str(r[RES_COL_TOTAL])
                finplace = ''
                if r[RES_COL_FINAL] >= 0:
                    finplace = str(r[RES_COL_FINAL] + 1)

                if self.scoring == 'madison':
                    laps = r[RES_COL_LAPS]
                    if fl is None:
                        fl = laps  # anddetermine laps down
                    if ll is not None:
                        down = fl - laps
                        if ll != laps:
                            sec.lines.append([
                                None, None,
                                unicode(down) + u' Lap' + strops.plural(down) +
                                u' Behind', None, None, None
                            ])
                    ll = laps
                if plstr or finplace or ptstr:  # only output those with points
                    # dnf  or
                    # placed in final sprint
                    sec.lines.append([plstr, rno, rname, rcat, fs, ptstr])
                    ## TEAM HACKS
                    if u't' in self.series and rh is not None:
                        for trno in rh[u'note'].split():
                            trh = self.meet.newgetrider(trno)  #!! SERIES?
                            if trh is not None:
                                trname = trh[u'namestr']
                                trinf = trh[u'ucicode']
                                sec.lines.append([
                                    None, u'', trname, trinf, None, None, None
                                ])
                fs = ''

        if self.onestart:
            sec.subheading = substr + u' - ' + self.standingstr()
        else:
            if substr:
                sec.subheading = substr

        ret.append(sec)

        if len(self.comments) > 0:
            sec = report.bullet_text()
            sec.subheading = u'Decisions of the commisaires panel'
            for c in self.comments:
                sec.lines.append([None, c])
            ret.append(sec)

        # output intermediate sprints?

        return ret

    def getrider(self, bib):
        """Return temporary reference to model row."""
        ret = None
        for r in self.riders:
            if r[RES_COL_BIB] == bib:
                ret = r
                break
        return ret

    def getiter(self, bib):
        """Return temporary iterator to model row."""
        i = self.riders.get_iter_first()
        while i is not None:
            if self.riders.get_value(i, RES_COL_BIB) == bib:
                break
            i = self.riders.iter_next(i)
        return i

    def addrider(self, bib='', info=None):
        """Add specified rider to race model."""
        nr = [bib, '', '', '', True, 0, 0, 0, '', -1, '', 0]
        er = self.getrider(bib)
        if bib == '' or er is None:
            dbr = self.meet.rdb.getrider(bib, self.series)
            if dbr is not None:
                for i in range(1, 4):
                    nr[i] = self.meet.rdb.getvalue(dbr, i)
                if self.inomnium:
                    if info:
                        nr[RES_COL_INFO] = info
            return self.riders.append(nr)
        else:
            if er is not None:
                #_log.debug('onestart is: %r', self.onestart)
                if self.inomnium and not self.onestart:
                    er[RES_COL_INFO] = info
            return None

    def delrider(self, bib):
        """Remove the specified rider from the model."""
        i = self.getiter(bib)
        if i is not None:
            self.riders.remove(i)

    def resettimer(self):
        """Reset race timer."""
        self.set_finish()
        self.set_start()
        self.timerstat = 'idle'
        self.meet.main_timer.dearm(0)
        self.meet.main_timer.dearm(1)
        self.stat_but.buttonchg(uiutil.bg_none, u'Idle')
        self.stat_but.set_sensitive(True)
        self.set_elapsed()

    def armstart(self):
        """Toggle timer arm start state."""
        if self.timerstat == 'idle':
            self.timerstat = 'armstart'
            self.stat_but.buttonchg(uiutil.bg_armstart, u'Arm Start')
            self.meet.main_timer.arm(0)
        elif self.timerstat == 'armstart':
            self.timerstat = 'idle'
            self.stat_but.buttonchg(uiutil.bg_none, u'Idle')
            self.meet.main_timer.dearm(0)
            self.curtimerstr = ''
        elif self.timerstat == 'running':
            self.timerstat = 'armsprintstart'
            self.stat_but.buttonchg(uiutil.bg_armstart, u'Arm Sprint')
            self.meet.main_timer.arm(0)
        elif self.timerstat == 'armsprintstart':
            self.timerstat = 'running'
            self.stat_but.buttonchg(uiutil.bg_none, u'Running')
            self.meet.main_timer.dearm(0)

    def armfinish(self):
        """Toggle timer arm finish state."""
        if self.timerstat in ['running', 'armsprint', 'armsprintstart']:
            self.timerstat = 'armfinish'
            self.stat_but.buttonchg(uiutil.bg_armfin, u'Arm Finish')
            self.meet.main_timer.arm(1)
        elif self.timerstat == 'armfinish':
            self.timerstat = 'running'
            self.stat_but.buttonchg(uiutil.bg_none, u'Running')
            self.meet.main_timer.dearm(1)

    def sort_handicap(self, x, y):
        """Sort by ranking, then info, then riderno"""
        if x[3] == y[3]:
            if x[2] != y[2]:
                if x[2] is None:  # y sorts first
                    return 1
                elif y[2] is None:  # x sorts first
                    return -1
                else:  # Both should be ints here
                    return cmp(x[2], y[2])
            else:  # Defer to rider number
                if x[1].isdigit() and y[1].isdigit():
                    return cmp(int(x[1]), int(y[1]))
                else:
                    return cmp(x[1], y[1])
        else:
            return cmp(x[3], y[3])

    def reorder_riderno(self):
        """Sort the rider list by rider number."""
        if len(self.riders) > 1:
            auxmap = []
            cnt = 0
            intmark = 0
            for r in self.riders:
                rno = r[RES_COL_BIB].decode(u'utf-8')
                seed = strops.confopt_posint(r[RES_COL_INFO].decode(u'utf-8'),
                                             9999)
                rank = 0
                if self.inomnium and self.evtype == u'omnium':
                    # extract rank from current standing
                    if r[RES_COL_PLACE].isdigit() or r[RES_COL_PLACE] == u'':
                        # but only add riders currently ranked, or unplaced
                        rank = strops.confopt_posint(
                            r[RES_COL_PLACE].decode(u'utf-8'), 9998)
                    else:
                        rank = 9999
                auxmap.append([cnt, rno, seed, rank])
                cnt += 1
            auxmap.sort(self.sort_handicap)
            #_log.debug('auxmap looks like: %r', auxmap)
            self.riders.reorder([a[0] for a in auxmap])

    def startlist_report(self, program=False):
        """Return a startlist report."""
        ret = []
        sec = report.twocol_startlist()
        if self.evtype == 'madison':
            # use the team singlecol method
            sec = report.section()
        headvec = [
            u'Event', self.evno, u':', self.event[u'pref'], self.event[u'info']
        ]
        if not program:
            headvec.append(u'- Start List')
        sec.heading = u' '.join(headvec)
        lapstring = strops.lapstring(self.event[u'laps'])
        substr = u' '.join(
            [lapstring, self.event[u'dist'], self.event[u'prog']]).strip()
        if substr:
            sec.subheading = substr
        self.reorder_riderno()
        sec.lines = []
        col2 = []
        cnt = 0
        if self.inomnium and len(self.riders) > 0:
            sec.lines.append([u' ', u' ', u'The Fence', None, None, None])
            col2.append([u' ', u' ', u'Sprinters Lane', None, None, None])
        for r in self.riders:
            rno = r[RES_COL_BIB].decode('utf-8')
            rh = self.meet.newgetrider(rno, self.series)
            inf = r[RES_COL_INFO].decode('utf-8', 'replace')
            if self.inomnium:
                # inf holds seed, ignore for now
                inf = None
            # layout needs adhjustment
            #if rh[u'ucicode']:
            #inf = rh[u'ucicode']   # overwrite by force
            rname = u''
            if rh is not None:
                rname = rh[u'namestr']

            if self.inomnium:
                if r[RES_COL_PLACE].isdigit() or r[RES_COL_PLACE] == u'':
                    cnt += 1
                    if cnt % 2 == 1:
                        sec.lines.append([None, rno, rname, inf, None, None])
                    else:
                        col2.append([None, rno, rname, inf, None, None])
            else:
                cnt += 1
                sec.lines.append([None, rno, rname, inf, None, None])
                if self.evtype == 'madison':
                    # add the black/red entry
                    if rh is not None:
                        tvec = rh[u'note'].split()
                        if len(tvec) == 2:
                            trname = u''
                            trinf = u''
                            trh = self.meet.newgetrider(tvec[0])
                            if trh is not None:
                                trname = trh[u'namestr']
                                trinf = trh[u'ucicode']
                            sec.lines.append([
                                None, u'Red', trname, trinf, None, None, None
                            ])
                            trname = u''
                            trinf = u''
                            trh = self.meet.newgetrider(tvec[1])
                            if trh is not None:
                                trname = trh[u'namestr']
                                trinf = trh[u'ucicode']
                            sec.lines.append([
                                None, u'Black', trname, trinf, None, None, None
                            ])
                            #sec.lines.append([None, None, None, None,
                            #None, None, None])

        for i in col2:
            sec.lines.append(i)
        ret.append(sec)

        # placeholders - why was this suppressed?
        if self.event[u'plac']:
            while cnt < self.event[u'plac']:
                sec.lines.append([None, None, None, None, None, None])
                cnt += 1

        if cnt > 0 and not program:
            sec = report.bullet_text()
            if self.evtype == 'madison':
                sec.lines.append([None, 'Total teams: ' + str(cnt)])
            else:
                sec.lines.append([None, 'Total riders: ' + str(cnt)])
            ret.append(sec)
        return ret

    def do_startlist(self):
        """Show startlist on scoreboard."""
        self.reorder_riderno()
        self.meet.scbwin = None
        self.timerwin = False
        startlist = []
        name_w = self.meet.scb.linelen - 8
        for r in self.riders:
            if r[RES_COL_INRACE]:
                nfo = r[RES_COL_CLUB].decode(u'utf-8')
                if self.inomnium:
                    if self.evtype == u'omnium':
                        # this is the omnium aggregate, use standings for nfo
                        nfo = r[RES_COL_PLACE].decode(u'utf-8')
                    else:
                        # overwrite nfo with seed value
                        nfo = r[RES_COL_INFO].decode(u'utf-8')
                else:
                    if len(nfo) > 3:
                        nfo = nfo[0:3]
                        ## look it up?
                        #if self.series in self.meet.ridermap:
                        #rh = self.meet.ridermap[self.series][
                        #r[RES_COL_BIB]]
                        #if rh is not None:
                        #nfo = rh['note']
                startlist.append([
                    r[RES_COL_BIB],
                    strops.fitname(r[RES_COL_FIRST].decode('utf-8'),
                                   r[RES_COL_LAST].decode('utf-8'), name_w),
                    nfo
                ])
        FMT = [(3, u'r'), u' ', (name_w, u'l'), u' ', (3, u'r')]
        self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                           head=self.meet.racenamecat(
                                               self.event),
                                           subhead=u'STARTLIST',
                                           coldesc=FMT,
                                           rows=startlist)
        self.meet.scbwin.reset()

    def key_event(self, widget, event):
        """Race window key press handler."""
        if event.type == gtk.gdk.KEY_PRESS:
            key = gtk.gdk.keyval_name(event.keyval) or 'None'
            if event.state & gtk.gdk.CONTROL_MASK:
                if key == key_abort:  # override ctrl+f5
                    self.resettimer()
                    return True
            if key[0] == 'F':
                if key == key_armstart:
                    self.armstart()
                    return True
                elif key == key_armfinish:
                    self.armfinish()
                    return True
                elif key == key_showtimer:
                    self.showtimer()
                    return True
                elif key == key_startlist:
                    self.do_startlist()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_results:
                    self.recalculate()
                    self.do_places()
                    glib.idle_add(self.delayed_announce)
                    return True
                elif key == key_lapdown:
                    if self.runlap is not None and self.runlap > 0:
                        self.runlap -= 1
                    return True
        return False

    def delayed_announce(self):
        """Initialise the announcer's screen after a delay."""
        if self.winopen:
            self.meet.txt_clear()
            self.meet.txt_title(u' '.join([
                self.meet.event_string(self.evno), u':', self.event[u'pref'],
                self.event[u'info']
            ]))

            self.meet.txt_line(1, u'_')
            self.meet.txt_line(8, u'_')

            # fill in a sprint if not empty
            sid = None
            i = self.ctrl_place_combo.get_active_iter()
            if i is not None:
                pl = self.sprints.get_value(i,
                                            SPRINT_COL_PLACES).decode(u'utf-8')
                if pl is not None and pl != '':
                    sinfo = self.sprints.get_value(
                        i, SPRINT_COL_LABEL).decode(u'utf-8')
                    self.meet.txt_setline(3, sinfo + u':')
                    sid = int(self.sprints.get_string_from_iter(i))
                    cnt = 0
                    unitshown = False
                    for r in self.sprintresults[sid]:
                        pstr = u''
                        if r[3] != u'':
                            pstr = r[3]
                            if not unitshown:
                                pstr += u'pts'
                                unitshown = True
                        self.meet.txt_postxt(
                            4 + cnt, 0, u' '.join([
                                strops.truncpad(r[0], 3),
                                strops.truncpad(r[1], 3, u'r'),
                                strops.truncpad(r[2], 20), pstr
                            ]))
                        cnt += 1
                        if cnt > 3:  # is this required?
                            break
                else:
                    sid = int(self.sprints.get_string_from_iter(i)) - 1

            tp = u''
            if self.start is not None and self.finish is not None:
                et = self.finish - self.start
                tp = u'Time: ' + et.timestr(2) + u'    '
                dist = self.meet.get_distance(self.distance, self.units)
                if dist:
                    tp += u'Avg: ' + et.speedstr(dist)
            self.meet.txt_postxt(4, 40, tp)

            # do result standing
            mscount = len(self.sprints)
            if sid is not None:
                mscount = sid
            sidstart = mscount - 8
            if sidstart < 0:
                sidstart = 0
            elif sidstart > len(self.sprints) - 10:
                sidstart = len(self.sprints) - 10
            if self.scoring == u'madison':
                leaderboard = []
                rtype = u'Team '
                if self.evtype != u'madison':
                    rtype = u'Rider'
                hdr = u'     # ' + rtype + u'                 Lap Pt '
                nopts = u''
                scnt = 0
                for s in self.sprints:
                    if scnt >= sidstart and scnt < sidstart + 10:
                        hdr += strops.truncpad(s[SPRINT_COL_ID], 4, u'r')
                        nopts += u'    '
                    scnt += 1
                hdr += u' Fin'
                self.meet.txt_setline(10, hdr)
                curline = 11
                ldrlap = None
                curlap = None
                for r in self.riders:
                    if ldrlap is None:
                        ldrlap = r[RES_COL_LAPS]
                        curlap = ldrlap
                    lapdwn = r[RES_COL_LAPS] - ldrlap
                    lapstr = u'  '
                    if lapdwn != 0:
                        lapstr = strops.truncpad(unicode(lapdwn), 2, u'r')

                    psrc = u'-'
                    if r[RES_COL_TOTAL] != 0:
                        psrc = unicode(r[RES_COL_TOTAL])
                    ptstr = strops.truncpad(psrc, 2, u'r')

                    placestr = u'   '
                    if self.onestart and r[RES_COL_PLACE] != u'':
                        placestr = strops.truncpad(r[RES_COL_PLACE] + u'.', 3)
                    elif not r[RES_COL_INRACE]:
                        placestr = u'dnf'

                    spstr = u''
                    if r[RES_COL_BIB] in self.auxmap:
                        scnt = 0
                        for s in self.auxmap[r[RES_COL_BIB]]:
                            if scnt >= sidstart and scnt < sidstart + 10:
                                spstr += str(s).rjust(4)
                            scnt += 1
                    else:
                        spstr = nopts

                    finstr = u'u/p'
                    if r[RES_COL_FINAL] >= 0:
                        finstr = strops.truncpad(str(r[RES_COL_FINAL] + 1), 3,
                                                 u'r')

                    bibstr = strops.truncpad(r[RES_COL_BIB], 2, u'r')

                    clubstr = u''
                    if r[RES_COL_CLUB] != u'' and len(r[RES_COL_CLUB]) <= 3:
                        clubstr = u' (' + r[RES_COL_CLUB] + u')'
                    namestr = strops.truncpad(strops.fitname(r[RES_COL_FIRST],
                                                             r[RES_COL_LAST],
                                                             22 - len(clubstr),
                                                             trunc=True) +
                                              clubstr,
                                              22,
                                              ellipsis=False)

                    self.meet.txt_postxt(
                        curline, 0, u' '.join([
                            placestr, bibstr, namestr, lapstr, ptstr, spstr,
                            finstr
                        ]))
                    curline += 1
                    if r[RES_COL_INRACE]:
                        if curlap > r[RES_COL_LAPS]:
                            while curlap != r[RES_COL_LAPS]:
                                curlap -= 1
                                if curlap < -15:
                                    break
                                leaderboard.append(u'-')
                        leaderboard.append(r[RES_COL_BIB].rjust(2) +
                                           psrc.rjust(3))
                self.meet.cmd_announce(u'leaderboard',
                                       unichr(unt4.US).join(leaderboard))
            else:
                # use scratch race style layout for up to 26 riders
                count = 0
                curline = 11
                posoft = 0
                leaderboard = []
                for r in self.riders:
                    count += 1
                    if count == 14:
                        curline = 11
                        posoft = 41

                    psrc = u'-'
                    if r[RES_COL_TOTAL] != 0:
                        psrc = unicode(r[RES_COL_TOTAL])

                    ptstr = strops.truncpad(psrc, 3, u'r')
                    clubstr = u''
                    if r[RES_COL_CLUB] != u'' and len(r[RES_COL_CLUB]) <= 3:
                        clubstr = u' (' + r[RES_COL_CLUB] + u')'
                    namestr = strops.truncpad(strops.fitname(r[RES_COL_FIRST],
                                                             r[RES_COL_LAST],
                                                             27 - len(clubstr),
                                                             trunc=True) +
                                              clubstr,
                                              27,
                                              ellipsis=False)
                    placestr = u'   '
                    if self.onestart and r[RES_COL_PLACE] != u'':
                        placestr = strops.truncpad(r[RES_COL_PLACE] + u'.', 3)
                    elif not r[RES_COL_INRACE]:
                        placestr = u'dnf'
                    bibstr = strops.truncpad(r[RES_COL_BIB], 3, u'r')
                    self.meet.txt_postxt(
                        curline, posoft,
                        u' '.join([placestr, bibstr, namestr, ptstr]))
                    curline += 1

                    if self.inomnium and r[RES_COL_INRACE]:
                        leaderboard.append(r[RES_COL_BIB].rjust(2) +
                                           psrc.rjust(3))
                if posoft > 0:
                    self.meet.txt_postxt(
                        10, 0,
                        u'      # Rider                       Pts        # Rider                       Pts'
                    )
                else:
                    self.meet.txt_postxt(
                        10, 0, u'      # Rider                       Pts')

                self.meet.cmd_announce(u'leaderboard',
                                       unichr(unt4.US).join(leaderboard))
        return False

    def do_places(self):
        """Show race result on scoreboard."""

        thesec = self.result_report()
        placestype = u'Standings: '
        if len(thesec) > 0:
            if self.finished:
                placestype = u'Result: '
        resvec = []
        fmt = ''
        hdr = ''
        if self.scoring == 'madison':
            name_w = self.meet.scb.linelen - 8
            fmt = [(2, u'r'), u' ', (name_w, u'l'), (2, u'r'), (3, u'r')]
            # does this require special consideration?
            leaderboard = []
            hdr = u' # team' + (
                (self.meet.scb.linelen - 13) * u' ') + u'lap pt'
            llap = None  # leader's lap
            for r in self.riders:
                if r[RES_COL_INRACE]:
                    bstr = r[RES_COL_BIB].decode('utf-8', 'ignore')
                    if llap is None:
                        llap = r[RES_COL_LAPS]
                    lstr = str(r[RES_COL_LAPS] - llap)
                    if lstr == '0': lstr = ''
                    pstr = str(r[RES_COL_TOTAL])
                    if pstr == '0': pstr = '-'
                    resvec.append([
                        bstr,
                        strops.fitname(
                            '', r[RES_COL_LAST].decode('utf-8',
                                                       'ignore').upper(),
                            name_w), lstr, pstr
                    ])
        else:
            name_w = self.meet.scb.linelen - 10
            fmt = [(3, u'l'), (3, u'r'), u' ', (name_w, u'l'), (3, u'r')]
            #self.meet.scb.linelen - 3) + ' pt'
            if self.inomnium:
                name_w -= 1
                fmt = [(3, u'l'), (3, u'r'), u' ', (name_w, u'l'), (4, u'r')]
                #self.meet.scb.linelen - 3) + ' pt'
            #ldr = None
            for r in self.riders:
                if r[RES_COL_INRACE]:
                    plstr = r[RES_COL_PLACE] + u'.'
                    bstr = r[RES_COL_BIB]
                    #if ldr is None and r[RES_COL_TOTAL] > 0:
                    #ldr = r[RES_COL_TOTAL]	# current leader
                    pstr = str(r[RES_COL_TOTAL])
                    #if self.inomnium and ldr is not None:
                    #pstr = '-' + str(ldr - r[RES_COL_TOTAL])
                    if pstr == '0': pstr = '-'
                    resvec.append([
                        plstr, bstr,
                        strops.fitname(r[RES_COL_FIRST].decode(u'utf-8'),
                                       r[RES_COL_LAST].decode(u'utf-8'),
                                       name_w), pstr
                    ])
            # cols are: rank, bib, name, pts
        hdr = self.meet.racenamecat(self.event)
        self.meet.scbwin = None
        self.timerwin = False
        evtstatus = self.standingstr(width=self.meet.scb.linelen).upper()
        #evtstatus=u'STANDINGS'
        #if self.finished:
        #evtstatus=u'RESULT'
        self.meet.scbwin = scbwin.scbtable(scb=self.meet.scb,
                                           head=hdr,
                                           subhead=evtstatus,
                                           coldesc=fmt,
                                           rows=resvec)
        self.meet.scbwin.reset()
        return False

    def dnfriders(self, biblist=''):
        """Remove listed bibs from the race."""
        recalc = False
        for bib in biblist.split():
            r = self.getrider(bib)
            if r is not None:
                r[RES_COL_INRACE] = False
                recalc = True
                _log.info(u'Rider %r withdrawn', bib)
            else:
                _log.warning('Did not withdraw %r', bib)
        if recalc:
            self.recalculate()
            self.meet.delayed_export()
        return False

    def announce_packet(self, line, pos, txt):
        self.meet.txt_postxt(line, pos, txt)
        return False

    def gainlap(self, biblist=u''):
        """Credit each rider listed in biblist with a lap on the field."""
        recalc = False
        rlines = []
        srlines = []
        for bib in biblist.split():
            r = self.getrider(bib)
            if r is not None:
                r[RES_COL_LAPS] += 1
                recalc = True
                _log.info(u'Rider %r gains a lap', bib)
                rlines.append(u' '.join([
                    bib.rjust(3),
                    strops.fitname(r[RES_COL_FIRST].decode(u'utf-8'),
                                   r[RES_COL_LAST].decode(u'utf-8'),
                                   26,
                                   trunc=True)
                ]))
                srlines.append([
                    bib,
                    strops.fitname(r[RES_COL_FIRST].decode(u'utf-8'),
                                   r[RES_COL_LAST].decode(u'utf-8'), 20)
                ])
            else:
                _log.warning(u'Did not gain lap for %r', bib)
        if recalc:
            self.oktochangecombo = False
            self.recalculate()
            glib.timeout_add_seconds(2, self.announce_packet, 3, 50,
                                     u'Gaining a lap:')
            cnt = 1
            for line in rlines:
                glib.timeout_add_seconds(2, self.announce_packet, 3 + cnt, 50,
                                         line)
                cnt += 1
                if cnt > 4:
                    break
            # and do it on the scoreboard too ?!
            self.meet.scbwin = scbwin.scbintsprint(
                self.meet.scb, self.meet.racenamecat(self.event),
                u'Gaining a Lap'.upper(), [(3, u'r'), u' ',
                                           (self.meet.scb.linelen - 4, u'l')],
                srlines)
            self.meet.scbwin.reset()
            self.next_sprint_counter += 1
            delay = SPRINT_PLACE_DELAY * len(rlines) or 1
            if delay > SPRINT_PLACE_DELAY_MAX:
                delay = SPRINT_PLACE_DELAY_MAX
            glib.timeout_add_seconds(delay, self.delayed_result)
            self.meet.delayed_export()
        return False

    def loselap(self, biblist=u''):
        """Deduct a lap from each rider listed in biblist."""
        recalc = False
        rlines = []
        for bib in biblist.split():
            r = self.getrider(bib)
            if r is not None:
                r[RES_COL_LAPS] -= 1
                recalc = True
                _log.info(u'Rider %r loses a lap', bib)
                rlines.append(u' '.join([
                    bib.rjust(3),
                    strops.fitname(r[RES_COL_FIRST].decode(u'utf-8'),
                                   r[RES_COL_LAST].decode(u'utf-8'),
                                   26,
                                   trunc=True)
                ]))
            else:
                _log.warning(u'Did not lose lap for %r', bib)
        if recalc:
            self.oktochangecombo = False
            self.recalculate()
            glib.timeout_add_seconds(2, self.announce_packet, 3, 50,
                                     u'Losing a lap:')
            cnt = 1
            for line in rlines:
                glib.timeout_add_seconds(2, self.announce_packet, 3 + cnt, 50,
                                         line)
                cnt += 1
                if cnt > 4:
                    break
            self.meet.delayed_export()
        return False

    def showtimer(self):
        """Show race timer on scoreboard"""
        tp = u'Time:'
        self.meet.scbwin = scbwin.scbtimer(scb=self.meet.scb,
                                           line1=self.meet.racenamecat(
                                               self.event),
                                           line2=u'',
                                           timepfx=tp)
        wastimer = self.timerwin
        self.timerwin = True
        if self.timerstat == 'finished':
            if not wastimer:
                self.meet.scbwin.reset()
            elap = self.finish - self.start
            self.meet.scbwin.settime(elap.timestr(2))
            dist = self.meet.get_distance(self.distance, self.units)
            if dist:
                self.meet.scbwin.setavg(elap.speedstr(dist))
                self.meet.scbwin.update()
        else:
            self.meet.scbwin.reset()

    def shutdown(self, win=None, msg=u'Exiting'):
        """Terminate event object"""
        _log.debug('Event shutdown: %r', msg)
        if not self.readonly:
            self.saveconfig()
        self.winopen = False

    def starttrig(self, e):
        """React to start trigger."""
        if self.timerstat == 'armstart':
            if self.distance and self.units == u'laps':
                self.runlap = self.distance - 1
                _log.debug(u'SET RUNLAP: %r', self.runlap)
            self.set_start(e, tod.now())
        elif self.timerstat == 'armsprintstart':
            self.stat_but.buttonchg(uiutil.bg_armfin, u'Arm Sprint')
            self.meet.main_timer.arm(1)
            self.timerstat = 'armsprint'
            self.sprintstart = e
            self.sprintlstart = tod.now()

    def fintrig(self, e):
        """React to finish trigger"""
        if self.timerstat == 'armfinish':
            self.set_finish(e)
            self.set_elapsed()
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                self.showtimer()
            self.log_elapsed()
            glib.idle_add(self.delayed_announce)
        elif self.timerstat == 'armsprint':
            self.stat_but.buttonchg(uiutil.bg_none, u'Running')
            self.timerstat = 'running'
            if self.sprintstart is not None:
                elap = (e - self.sprintstart).timestr(2)
                _log.info(u'200m: %s', elap)
                if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                    self.meet.scbwin.avgpfx = u'200m:'
                    self.meet.scbwin.setavg(elap)
            self.sprintstart = None

    def timercb(self, e):
        """Handle a timer event"""
        chan = timy.chan2id(e.chan)
        if chan == 0:
            _log.debug(u'Start impulse %s', e.rawtime(3))
            self.starttrig(e)
        elif chan == 1:
            _log.debug(u'Finish impulse %s', e.rawtime(3))
            self.fintrig(e)
        return False

    def timeout(self):
        """Update scoreboard and respond to timing events"""
        if not self.winopen:
            return False
        if self.finish is None and self.start is not None:
            self.set_elapsed()
            if self.timerwin and type(self.meet.scbwin) is scbwin.scbtimer:
                self.meet.scbwin.settime(
                    self.time_lbl.get_text().decode(u'utf-8'))
        return True

    def do_properties(self):
        """Run race properties dialog"""
        b = gtk.Builder()
        b.add_from_file(os.path.join(metarace.UI_PATH, u'ps_properties.ui'))
        dlg = b.get_object(u'properties')
        dlg.set_transient_for(self.meet.window)
        rle = b.get_object(u'race_laps_entry')
        rle.set_text(self.sprintlaps)
        if self.onestart:
            rle.set_sensitive(False)
        rsb = b.get_object(u'race_showbib_toggle')
        rsb.set_active(self.masterslaps)
        rt = b.get_object(u'race_score_type')
        if self.scoring == u'madison':
            rt.set_active(0)
        else:
            rt.set_active(1)
        di = b.get_object(u'race_dist_entry')
        if self.distance is not None:
            di.set_text(unicode(self.distance))
        else:
            di.set_text(u'')
        du = b.get_object(u'race_dist_type')
        if self.units == u'metres':
            du.set_active(0)
        else:
            du.set_active(1)
        se = b.get_object(u'race_series_entry')
        se.set_text(self.series)
        as_e = b.get_object(u'auto_starters_entry')
        as_e.set_text(self.autospec)

        response = dlg.run()
        if response == 1:  # id 1 set in glade for "Apply"
            _log.debug(u'Updating race properties')
            if not self.onestart:
                newlaps = strops.reformat_biblist(
                    rle.get_text().decode(u'utf-8'))
                if self.sprintlaps != newlaps:
                    self.sprintlaps = newlaps
                    _log.info(u'Reset sprint model')
                    self.sprint_model_init()
            self.masterslaps = rsb.get_active()
            if rt.get_active() == 0:
                self.scoring = u'madison'
            else:
                self.scoring = u'points'
            self.type_lbl.set_text(self.scoring.capitalize())
            dval = di.get_text().decode(u'utf-8')
            if dval.isdigit():
                self.distance = int(dval)
            else:
                self.distance = None
            if du.get_active() == 0:
                self.units = u'metres'
            else:
                self.units = u'laps'

            # update series
            ns = se.get_text().decode(u'utf-8')
            if ns != self.series:
                self.series = ns
                self.event[u'seri'] = ns

            # update auto startlist spec
            nspec = as_e.get_text().decode(u'utf-8')
            if nspec != self.autospec:
                self.autospec = nspec
                if not self.onestart:
                    if self.autospec:
                        self.meet.autostart_riders(self, self.autospec,
                                                   self.seedsrc)

            # xfer starters if not empty
            slist = strops.riderlist_split(
                b.get_object(u'race_starters_entry').get_text().decode(
                    u'utf-8'), self.meet.rdb, self.series)
            for s in slist:
                self.addrider(s)

            # recalculate
            self.reset_lappoints()
            self.recalculate()
            glib.idle_add(self.delayed_announce)
        else:
            _log.debug(u'Edit race properties cancelled')

        # if prefix is empty, grab input focus
        if not self.prefix_ent.get_text():
            self.prefix_ent.grab_focus()
        dlg.destroy()

    ## Race timing manipulations
    def set_start(self, start=u'', lstart=None):
        """Set the race start time."""
        if type(start) is tod.tod:
            self.start = start
            if lstart is not None:
                self.lstart = lstart
            else:
                self.lstart = self.start
        else:
            self.start = tod.mktod(start)
            if lstart is not None:
                self.lstart = tod.mktod(lstart)
            else:
                self.lstart = self.start
        if self.start is None:
            pass
        else:
            if self.finish is None:
                self.set_running()

    def set_finish(self, finish=u''):
        """Set the race finish time."""
        if type(finish) is tod.tod:
            self.finish = finish
        else:
            self.finish = tod.mktod(finish)
        if self.finish is None:
            if self.start is not None:
                self.set_running()
        else:
            if self.start is None:
                self.set_start(tod.ZERO)
            self.set_finished()

    def set_elapsed(self):
        """Update elapsed race time."""
        if self.start is not None and self.finish is not None:
            et = self.finish - self.start
            self.time_lbl.set_text(et.timestr(2))
        elif self.start is not None:
            runtm = (tod.now() - self.lstart).timestr(1)
            self.time_lbl.set_text(runtm)

            if self.runlap is not None:
                if self.runlap != self.lastrunlap:
                    _log.debug(u'Runlap: %r', self.runlap)
                    self.lastrunlap = self.runlap

        elif self.timerstat == u'armstart':
            self.time_lbl.set_text(tod.tod(0).timestr(1))
            if self.runlap and self.runlap != self.lastrunlap:
                _log.debug(u'Runlap: %r', self.runlap)
                self.lastrunlap = self.runlap
        else:
            self.time_lbl.set_text(u'')

    def log_elapsed(self):
        """Log elapsed time on timy receipt"""
        self.meet.main_timer.printline(self.meet.racenamecat(self.event))
        self.meet.main_timer.printline(u'      ST: ' + self.start.timestr(4))
        self.meet.main_timer.printline(u'     FIN: ' + self.finish.timestr(4))
        self.meet.main_timer.printline(u'    TIME: ' +
                                       (self.finish - self.start).timestr(2))

    def set_running(self):
        """Set timer to running"""
        self.timerstat = u'running'
        self.stat_but.buttonchg(uiutil.bg_none, u'Running')

    def set_finished(self):
        """Set timer to finished"""
        self.timerstat = u'finished'
        self.stat_but.buttonchg(uiutil.bg_none, u'Finished')
        self.stat_but.set_sensitive(False)
        self.ctrl_places.grab_focus()

    def update_expander_lbl_cb(self):
        """Update the expander label"""
        self.info_expand.set_label(u'Race Info : ' +
                                   self.meet.racenamecat(self.event, 64))

    def ps_info_time_edit_clicked_cb(self, button, data=None):
        """Run the edit times dialog."""
        ostx = u''
        oftx = u''
        if self.start is not None:
            ostx = self.start.rawtime(4)
        if self.finish is not None:
            oftx = self.finish.rawtime(4)
        ret = uiutil.edit_times_dlg(self.meet.window, ostx, oftx)
        if ret[0] == 1:  # id 1 set in glade for "Apply"
            try:
                stod = None
                if ret[1]:
                    stod = tod.tod(ret[1], u'MANU', u'C0i')
                    self.meet.main_timer.printline(u' ' + str(stod))
                ftod = None
                if ret[2]:
                    ftod = tod.tod(ret[2], u'MANU', u'C1i')
                    self.meet.main_timer.printline(u' ' + str(ftod))
                self.set_start(stod)
                self.set_finish(ftod)
                self.set_elapsed()
                if self.start is not None and self.finish is not None:
                    self.log_elapsed()
                _log.info(u'Updated race times')
            except Exception as v:
                _log.error(u'%s updating times: %s', v.__class__.__name__, v)
        else:
            _log.info(u'Edit race times cancelled')

    def ps_ctrl_place_combo_changed_cb(self, combo, data=None):
        """Handle sprint combo change."""
        self.oktochangecombo = False  # cancel delayed combo changer
        i = self.ctrl_place_combo.get_active_iter()
        if i is not None:
            self.ctrl_places.set_text(
                self.sprints.get_value(i, SPRINT_COL_PLACES) or u'')
        else:
            self.ctrl_places.set_text(u'')
        self.ctrl_places.grab_focus()

    def standingstr(self, width=None):
        """Return an event status string for reports and scb."""
        ret = u'Standings'
        totsprints = 0
        lastsprint = None
        sprintid = None
        cur = 1
        for s in self.sprints:
            #_log.debug(u'cur: %r, val: %r', cur, s[SPRINT_COL_PLACES])
            if s[SPRINT_COL_PLACES]:
                lastsprint = cur
                sprintid = s[SPRINT_COL_ID]
            if s[SPRINT_COL_ID] not in self.laplabels:
                totsprints += 1
                cur += 1
        if lastsprint >= totsprints:
            ret = u'Result'
            # check for all places in final sprint?
            for r in self.riders:
                if r[RES_COL_INRACE] and r[RES_COL_FINAL] < 0:
                    # not placed at finish
                    ret = u'Provisional Result'
                    break
        else:
            ret = u'Standings'
            if lastsprint:
                if sprintid in self.laplabels:
                    ret += u' After ' + self.laplabels[sprintid]
                elif totsprints > 0:
                    if width is not None and width < 25:
                        ret += u' - Sprint {0}/{1}'.format(
                            lastsprint, totsprints)
                    else:
                        ret += u' After Sprint {0} of {1}'.format(
                            lastsprint, totsprints)
                else:
                    _log.debug(u'Total sprints was 0: %r / %r', lastsprint,
                               totsprints)
        return ret

    def delayed_result(self):
        """Roll the places entry on to the next sprint"""
        if self.next_sprint_counter > 1:
            self.next_sprint_counter -= 1
        elif self.next_sprint_counter == 1:
            self.next_sprint_counter = 0
            self.do_places()
            if self.ctrl_places.get_property('has-focus'):
                if self.oktochangecombo:
                    i = self.ctrl_place_combo.get_active_iter()
                    i = self.ctrl_place_combo.get_model().iter_next(i)
                    if i is not None:
                        self.ctrl_place_combo.set_active_iter(i)
            else:
                # input widget lost focus, don't auto advance
                self.oktochangecombo = False
        else:
            self.next_sprint_counter = 0  # clamp negatives
        return False

    def checkplaces(self, places=u''):
        """Check the proposed places against current race model"""
        ret = True
        placeset = set()
        for no in strops.reformat_biblist(places).split():
            # repetition? - already in place set?
            if no in placeset:
                _log.error(u'Duplicate no in places: %r', no)
                ret = False
            placeset.add(no)
            # rider in the model?
            lr = self.getrider(no)
            if lr is None:
                if not self.meet.get_clubmode():
                    _log.error(u'Non-starter in places: %r', no)
                    ret = False
                # otherwise club mode allows non-starter in places
            else:
                # rider still in the race?
                if not lr[RES_COL_INRACE]:
                    _log.error(u'DNF rider in places: %r', no)
                    ret = False
        return ret

    def ps_ctrl_places_activate_cb(self, entry, data=None):
        """Handle places entry"""
        places = strops.reformat_placelist(entry.get_text().decode(u'utf-8'))
        if self.checkplaces(places):
            self.oktochangecombo = False  # cancel existing delayed change
            entry.set_text(places)

            i = self.ctrl_place_combo.get_active_iter()
            prevplaces = self.sprints.get_value(
                i, SPRINT_COL_PLACES).decode(u'utf-8')
            self.sprints.set_value(i, SPRINT_COL_PLACES, places)
            sid = int(self.sprints.get_string_from_iter(i))
            sinfo = self.sprints.get_value(i,
                                           SPRINT_COL_LABEL).decode(u'utf-8')
            self.recalculate()
            self.timerwin = False
            _log.info('%s: %r', sinfo, places)
            if prevplaces == u'':
                FMT = [(2, u'l'), (3, u'r'), u' ',
                       (self.meet.scb.linelen - 8, u'l'), u' ', (1, u'r')]
                self.meet.scbwin = scbwin.scbintsprint(
                    self.meet.scb, self.meet.racenamecat(self.event),
                    sinfo.upper(), FMT, self.sprintresults[sid][0:4])
                self.meet.scbwin.reset()
                self.next_sprint_counter += 1
                delay = SPRINT_PLACE_DELAY * len(self.sprintresults[sid]) or 1
                if delay > SPRINT_PLACE_DELAY_MAX:
                    delay = SPRINT_PLACE_DELAY_MAX
                self.oktochangecombo = True
                glib.timeout_add_seconds(delay, self.delayed_result)
            elif type(self.meet.scbwin) is scbwin.scbtable:
                self.do_places()  # overwrite result table?
            glib.timeout_add_seconds(1, self.delayed_announce)
            self.meet.delayed_export()
        else:
            _log.error(u'Places not updated')

    def ps_ctrl_action_combo_changed_cb(self, combo, data=None):
        """Handle change on action combo."""
        self.ctrl_action.set_text(u'')
        self.ctrl_action.grab_focus()

    def ps_ctrl_action_activate_cb(self, entry, data=None):
        """Perform current action on bibs listed."""
        rlist = entry.get_text().decode(u'utf-8')
        acode = self.action_model.get_value(
            self.ctrl_action_combo.get_active_iter(), 1)
        if acode == 'gain':
            self.gainlap(strops.reformat_biblist(rlist))
            entry.set_text(u'')
        elif acode == 'lost':
            self.loselap(strops.reformat_biblist(rlist))
            entry.set_text(u'')
        elif acode == 'dnf':
            self.dnfriders(strops.reformat_biblist(rlist))
            entry.set_text(u'')
        elif acode == 'add':
            rlist = strops.riderlist_split(rlist, self.meet.rdb, self.series)
            for bib in rlist:
                self.addrider(bib)
            entry.set_text(u'')
        elif acode == 'del':
            rlist = strops.riderlist_split(rlist, self.meet.rdb, self.series)
            for bib in rlist:
                self.delrider(bib)
            entry.set_text(u'')
        elif acode == 'lap':
            self.runlap = strops.confopt_posint(rlist)
        else:
            _log.debug(u'Ignoring invalid action')
        glib.idle_add(self.delayed_announce)

    def ps_sprint_cr_label_edited_cb(self, cr, path, new_text, data=None):
        """Sprint column edit"""
        self.sprints[path][SPRINT_COL_LABEL] = new_text

    def ps_sprint_cr_places_edited_cb(self, cr, path, new_text, data=None):
        """Sprint place edit"""
        new_text = strops.reformat_placelist(new_text)
        self.sprints[path][SPRINT_COL_PLACES] = new_text
        opath = self.sprints.get_string_from_iter(
            self.ctrl_place_combo.get_active_iter())
        if opath == path:
            self.ctrl_places.set_text(new_text)
        self.recalculate()
        # edit places outside control - no auto trigger of export

    def editcol_db(self, cell, path, new_text, col):
        """Cell update with writeback to meet"""
        new_text = new_text.strip()
        self.riders[path][col] = new_text.strip()
        glib.idle_add(self.meet.rider_edit, self.riders[path][RES_COL_BIB],
                      self.series, col, new_text)

    def editcol_cb(self, cell, path, new_text, col):
        self.riders[path][col] = new_text.strip()

    def ps_result_cr_inrace_toggled_cb(self, cr, path, data=None):
        self.riders[path][RES_COL_INRACE] = not (
            self.riders[path][RES_COL_INRACE])
        self.recalculate()

    def ps_result_cr_laps_edited_cb(self, cr, path, new_text, data=None):
        try:
            laps = int(new_text)
            self.riders[path][RES_COL_LAPS] = laps
            self.recalculate()
        except ValueError:
            _log.warning(u'Ignoring non-numeric lap count')

    def zeropoints(self):
        for r in self.riders:
            r[RES_COL_POINTS] = 0
            r[RES_COL_TOTAL] = 0
            r[RES_COL_PLACE] = u''
            r[RES_COL_FINAL] = -1  # Negative => Unplaced in final sprint

    def pointsxfer(self, placestr, final=False, index=0, points=None):
        """Transfer points from sprint placings to aggregate."""
        placeset = set()
        if points is None:
            points = [5, 3, 2, 1]  # Default is four places
        self.sprintresults[index] = []
        place = 0
        count = 0
        name_w = self.meet.scb.linelen - 8
        for placegroup in placestr.split():
            for bib in placegroup.split(u'-'):
                if bib not in placeset:
                    placeset.add(bib)
                    r = self.getrider(bib)
                    if r is None:  # ensure rider exists at this point
                        _log.info(u'Adding non-starter: %r', bib)
                        self.addrider(bib)
                        r = self.getrider(bib)
                    ptsstr = u''
                    if place < len(points):
                        ptsstr = str(points[place])
                        r[RES_COL_POINTS] += points[place]
                        if bib not in self.auxmap:
                            self.auxmap[bib] = self.nopts[0:]
                        self.auxmap[bib][index] = str(points[place])
                    if final:
                        r[RES_COL_FINAL] = place
                        self.finished = True
                    plstr = unicode(place + 1) + u'.'
                    fname = r[RES_COL_FIRST].decode('utf-8')
                    lname = r[RES_COL_LAST].decode('utf-8')
                    club = r[RES_COL_CLUB].decode('utf-8')
                    if len(club) > 3:
                        club = club[0:3]
                        ## look it up?
                        #if self.series in self.meet.ridermap:
                        #rh = self.meet.ridermap[self.series][bib]
                        #if rh is not None:
                        #club = rh[u'note']
                    self.sprintresults[index].append([
                        plstr, r[RES_COL_BIB],
                        strops.fitname(fname, lname, name_w), ptsstr,
                        strops.resname(fname, lname, club)
                    ])
                    count += 1
                else:
                    _log.error(u'Ignoring duplicate no: %r', bib)
            place = count
        if count > 0:
            self.onestart = True

    def retotal(self, r):
        """Update totals"""
        if self.scoring == u'madison':
            r[RES_COL_TOTAL] = r[RES_COL_STPTS] + r[RES_COL_POINTS]
        else:
            r[RES_COL_TOTAL] = r[RES_COL_STPTS] + r[RES_COL_POINTS] + (
                self.lappoints * r[RES_COL_LAPS])

    # Sorting performed in-place on aux table with cols:
    #  0 INDEX		Index in main model
    #  1 BIB		Rider's bib
    #  2 INRACE		Bool rider still in race?
    #  3 LAPS		Rider's laps up/down
    #  4 TOTAL		Total points scored
    #  5 FINAL		Rider's place in final sprint (-1 for unplaced)

    # Point score sorting:
    # inrace / points / final sprint
    def sortpoints(self, x, y):
        if x[2] != y[2]:  # compare inrace
            if x[2]:
                return -1
            else:
                return 1
        else:  # defer to points
            return self.sortpointsonly(x, y)

    def sortpointsonly(self, x, y):
        if x[6] == y[6]:
            if x[4] > y[4]:
                return -1
            elif x[4] < y[4]:
                return 1
            else:  # defer to last sprint
                if x[5] == y[5]:
                    #_log.warning('Sort could not split riders.')
                    return 0  # places same - or both unplaced
                else:
                    xp = x[5]
                    if xp < 0: xp = 9999
                    yp = y[5]
                    if yp < 0: yp = 9999
                    return cmp(xp, yp)
        else:
            return (cmp(x[6], y[6]))

    def sortmadison(self, x, y):
        """Lap-based points (old-style Madison)"""
        if x[2] != y[2]:  # compare inrace
            if x[2]:
                return -1
            else:
                return 1
        else:  # defer to distance (laps)
            if x[3] > y[3]:
                return -1
            elif x[3] < y[3]:
                return 1
            else:  # defer to points / final sprint
                return self.sortpointsonly(x, y)

    def sort_riderno(self, x, y):
        return cmp(strops.riderno_key(x[1]), strops.riderno_key(y[1]))

    # result recalculation
    def recalculate(self):
        self.zeropoints()
        self.finished = False
        self.auxmap = {}
        idx = 0
        for s in self.sprints:
            self.pointsxfer(s[SPRINT_COL_PLACES], s[SPRINT_COL_ID] == u'0',
                            idx, s[SPRINT_COL_POINTS])
            idx += 1

        if len(self.riders) == 0:
            return

        auxtbl = []
        idx = 0
        for r in self.riders:
            ptotal = r[RES_COL_TOTAL]
            ranker = 0
            if not r[RES_COL_INRACE]:
                ptotal = 0
                if r[RES_COL_INFO] and r[RES_COL_INFO].isdigit():
                    ranker = int(r[RES_COL_INFO])
            self.retotal(r)
            auxtbl.append([
                idx, r[RES_COL_BIB], r[RES_COL_INRACE], r[RES_COL_LAPS],
                r[RES_COL_TOTAL], r[RES_COL_FINAL], ranker
            ])
            idx += 1
        if self.scoring == u'madison':
            auxtbl.sort(self.sortmadison)
        else:
            auxtbl.sort(self.sortpoints)
        self.riders.reorder([a[0] for a in auxtbl])
        place = 0
        idx = 0
        for r in self.riders:
            if r[RES_COL_INRACE]:
                if idx == 0:
                    place = 1
                else:
                    if self.scoring == u'madison':
                        if self.sortmadison(auxtbl[idx - 1], auxtbl[idx]) != 0:
                            place = idx + 1
                    else:
                        if self.sortpoints(auxtbl[idx - 1], auxtbl[idx]) != 0:
                            place = idx + 1
                r[RES_COL_PLACE] = unicode(place)
                idx += 1
            else:
                r[RES_COL_PLACE] = u'dnf'

    def sprint_model_init(self):
        """Initialise the sprint places model"""
        self.ctrl_place_combo.set_active(-1)
        self.ctrl_places.set_sensitive(False)
        self.sprints.clear()
        self.auxmap = {}
        self.nopts = []
        isone = False
        self.sprintresults = []
        for sl in self.sprintlaps.split():
            isone = True
            lt = sl
            if sl.isdigit():
                if int(sl) == 0:
                    lt = u'Final sprint'
                else:
                    lt = u'Sprint at ' + sl + u' to go'
            sp = None
            if sl in self.sprintpoints:
                nextp = []
                for nv in self.sprintpoints[sl].split():
                    if nv.isdigit():
                        nextp.append(int(nv))
                    else:
                        nextp = None
                        break
                sp = nextp
            nr = [sl, lt, None, None, u'', sp]
            self.sprints.append(nr)
            self.sprintresults.append([])
            self.nopts.append(u'')
        if isone:
            self.ctrl_place_combo.set_active(0)
            self.ctrl_places.set_sensitive(True)

    def spptsedit(self, cr, path, new_text, data=None):
        """Sprint points edit"""
        new_text = strops.reformat_biblist(new_text.decode('utf-8'))
        op = None
        nextp = []
        for nv in new_text.split():
            if nv.isdigit():
                nextp.append(int(nv))
            else:
                nextp = None
                break
        sid = self.sprints[path][SPRINT_COL_ID]
        if nextp is not None and len(nextp) > 0:
            self.sprintpoints[sid] = u' '.join(map(unicode, nextp))
            op = nextp
        else:
            if sid in self.sprintpoints:
                del self.sprintpoints[sid]
        self.sprints[path][SPRINT_COL_POINTS] = op
        self.recalculate()

    def spptsstr(self, col, cr, model, iter, data=None):
        """Format tod into text for listview"""
        pv = model.get_value(iter, SPRINT_COL_POINTS)
        if pv is not None and len(pv) > 0:
            cr.set_property('text', u', '.join(map(unicode, pv)))
        else:
            cr.set_property('text', u'')

    def todstr(self, col, cr, model, iter, data=None):
        """Format tod into text for listview"""
        st = model.get_value(iter, SPRINT_COL_200)
        ft = model.get_value(iter, SPRINT_COL_SPLIT)
        if st is not None and ft is not None:
            cr.set_property('text', (ft - st).timestr(2))
        else:
            cr.set_property('text', '')

    def reset_lappoints(self):
        """Update lap points allocation"""
        if self.masterslaps:
            self.lappoints = 10
        else:
            self.lappoints = 20

    def destroy(self):
        """Signal race shutdown"""
        self.frame.destroy()

    def show(self):
        """Show race window"""
        self.frame.show()

    def hide(self):
        """Hide race window"""
        self.frame.hide()

    def editent_cb(self, entry, col):
        """Shared event entry update callback"""
        if col == u'pref':
            self.event[u'pref'] = entry.get_text().decode('utf-8', 'replace')
        elif col == u'info':
            self.event[u'info'] = entry.get_text().decode('utf-8', 'replace')
        self.update_expander_lbl_cb()

    def __init__(self, meet, event, ui=True):
        """Constructor"""
        self.meet = meet
        self.event = event
        self.evno = event[u'evid']
        self.evtype = event[u'type']
        self.series = event[u'seri']
        self.configfile = meet.event_configfile(self.evno)

        self.readonly = not ui
        rstr = u''
        if self.readonly:
            rstr = u'readonly '
        _log.debug(u'Init %sevent %s', rstr, self.evno)

        # race property attributes
        self.comments = []
        self.masterslaps = True
        self.lappoints = 20
        self.scoring = u'points'
        self.distance = None
        self.units = u'laps'
        self.sprintlaps = u''
        self.sprintpoints = {}
        self.nopts = []
        self.sprintresults = []
        self.laplabels = {}
        self.sprintsource = {}
        self.auxmap = {}

        # race run time attributes
        self.onestart = False
        self.runlap = None
        self.lastrunlap = None

        self.start = None
        self.lstart = None
        self.finish = None
        self.winopen = ui
        self.timerwin = False
        self.timerstat = u'idle'
        self.curtimerstr = u''
        self.sprintstart = None
        self.sprintlstart = None
        self.next_sprint_counter = 0
        self.oktochangecombo = False
        self.autospec = u''
        self.finished = False
        self.inomnium = False
        self.seedsrc = None

        # data models
        self.sprints = gtk.ListStore(
            gobject.TYPE_STRING,  # ID = 0
            gobject.TYPE_STRING,  # LABEL = 1
            gobject.TYPE_PYOBJECT,  # 200 = 2
            gobject.TYPE_PYOBJECT,  # SPLITS = 3
            gobject.TYPE_STRING,  # PLACES = 4
            gobject.TYPE_PYOBJECT)  # POINTS = 5

        self.riders = gtk.ListStore(
            gobject.TYPE_STRING,  # BIB = 0
            gobject.TYPE_STRING,  # FIRST = 1
            gobject.TYPE_STRING,  # LAST = 2
            gobject.TYPE_STRING,  # CLUB = 3
            gobject.TYPE_BOOLEAN,  # INRACE = 4
            gobject.TYPE_INT,  # POINTS = 5
            gobject.TYPE_INT,  # LAPS = 6
            gobject.TYPE_INT,  # TOTAL = 7
            gobject.TYPE_STRING,  # PLACE = 8
            gobject.TYPE_INT,  # FINAL = 9
            gobject.TYPE_STRING,  # INFO = 10
            gobject.TYPE_INT)  # STPTS = 11

        uifile = os.path.join(metarace.UI_PATH, u'ps.ui')
        _log.debug(u'Building event interface from %r', uifile)
        b = gtk.Builder()
        b.add_from_file(uifile)

        self.frame = b.get_object(u'ps_vbox')
        self.frame.connect(u'destroy', self.shutdown)

        # info pane
        self.info_expand = b.get_object(u'info_expand')
        b.get_object(u'ps_info_evno').set_text(self.evno)
        self.showev = b.get_object(u'ps_info_evno_show')
        self.prefix_ent = b.get_object(u'ps_info_prefix')
        self.prefix_ent.connect(u'changed', self.editent_cb, u'pref')
        self.prefix_ent.set_text(self.event[u'pref'])
        self.info_ent = b.get_object(u'ps_info_title')
        self.info_ent.connect(u'changed', self.editent_cb, u'info')
        self.info_ent.set_text(self.event[u'info'])

        self.time_lbl = b.get_object(u'ps_info_time')
        self.time_lbl.modify_font(uiutil.MONOFONT)
        self.update_expander_lbl_cb()
        self.type_lbl = b.get_object(u'race_type')
        self.type_lbl.set_text(self.scoring.capitalize())

        # ctrl pane
        self.stat_but = uiutil.statbut(b.get_object(u'ps_ctrl_stat_but'))
        self.ctrl_place_combo = b.get_object(u'ps_ctrl_place_combo')
        self.ctrl_place_combo.set_model(self.sprints)
        self.ctrl_places = b.get_object(u'ps_ctrl_places')
        self.ctrl_action_combo = b.get_object(u'ps_ctrl_action_combo')
        self.ctrl_action = b.get_object(u'ps_ctrl_action')
        self.action_model = b.get_object(u'ps_action_model')

        if ui:
            # sprints pane
            t = gtk.TreeView(self.sprints)
            t.set_reorderable(True)
            t.set_enable_search(False)
            t.set_rules_hint(True)
            t.show()
            uiutil.mkviewcoltxt(t,
                                u'Sprint',
                                SPRINT_COL_LABEL,
                                self.ps_sprint_cr_label_edited_cb,
                                expand=True)
            #uiutil.mkviewcoltod(t, u'200m', cb=self.todstr)
            uiutil.mkviewcoltxt(t,
                                u'Places',
                                SPRINT_COL_PLACES,
                                self.ps_sprint_cr_places_edited_cb,
                                expand=True)
            uiutil.mkviewcoltod(t,
                                u'Points',
                                cb=self.spptsstr,
                                editcb=self.spptsedit)
            b.get_object(u'ps_sprint_win').add(t)

            # results pane
            t = gtk.TreeView(self.riders)
            t.set_reorderable(True)
            t.set_enable_search(False)
            t.set_rules_hint(True)
            t.show()
            uiutil.mkviewcoltxt(t, u'No.', RES_COL_BIB, calign=1.0)
            uiutil.mkviewcoltxt(t,
                                u'First Name',
                                RES_COL_FIRST,
                                self.editcol_db,
                                expand=True)
            uiutil.mkviewcoltxt(t,
                                u'Last Name',
                                RES_COL_LAST,
                                self.editcol_db,
                                expand=True)
            uiutil.mkviewcoltxt(t, u'Club', RES_COL_CLUB, self.editcol_db)
            uiutil.mkviewcoltxt(t, u'Info', RES_COL_INFO, self.editcol_cb)
            uiutil.mkviewcolbool(t,
                                 u'In',
                                 RES_COL_INRACE,
                                 self.ps_result_cr_inrace_toggled_cb,
                                 width=50)
            uiutil.mkviewcoltxt(t,
                                u'Pts',
                                RES_COL_POINTS,
                                calign=1.0,
                                width=50)
            uiutil.mkviewcoltxt(t,
                                u'Laps',
                                RES_COL_LAPS,
                                calign=1.0,
                                width=50,
                                cb=self.ps_result_cr_laps_edited_cb)
            uiutil.mkviewcoltxt(t,
                                u'Total',
                                RES_COL_TOTAL,
                                calign=1.0,
                                width=50)
            uiutil.mkviewcoltxt(t, u'L/S', RES_COL_FINAL, calign=0.5, width=50)
            uiutil.mkviewcoltxt(t,
                                u'Rank',
                                RES_COL_PLACE,
                                calign=0.5,
                                width=50)
            b.get_object(u'ps_result_win').add(t)

            # connect signal handlers
            b.connect_signals(self)



"""Custom log handlers."""

import logging
import glib


class traceFilter(logging.Filter):
    """Filter events to type TIMER only."""

    def filter(self, record):
        return record.levelno == 25  # TIMER LOG LEVEL


class traceHandler(logging.Handler):
    """Class for capturing timer log traces."""

    def __init__(self, trace=None):
        self.__trace = trace
        logging.Handler.__init__(self)
        self.addFilter(traceFilter())

    def emit(self, record):
        """Append log record to trace."""
        if self.__trace is not None:
            msg = self.format(record)
            self.__trace.append(msg)


class textViewHandler(logging.Handler):
    """A class for displaying log messages in a GTK text view."""

    def __init__(self, log=None, view=None, scroll=None):
        self.log = log
        self.view = view
        self.scroll = scroll
        self.scroll_pending = False
        logging.Handler.__init__(self)

    def do_scroll(self):
        """Catchup end of scrolled window."""
        if self.scroll_pending:
            self.view.scroll_to_iter(self.log.get_end_iter(), 0)
            self.scroll_pending = False
        return False

    def append_log(self, msg):
        """Append msg to the text view."""
        atend = True
        if self.scroll and self.scroll.page_size > 0:
            # Fudge a 'sticky' end of scroll mode... about a pagesz
            if self.scroll.upper - (self.scroll.value + self.scroll.page_size
                                    ) > (0.5 * self.scroll.page_size):
                atend = False
        self.log.insert(self.log.get_end_iter(), msg.strip() + '\n')
        if atend:
            self.scroll_pending = True
            glib.timeout_add_seconds(1, self.do_scroll)
        return False

    def emit(self, record):
        """Emit log record to gtk main loop."""
        msg = self.format(record)
        glib.idle_add(self.append_log, msg)


class statusHandler(logging.Handler):
    """A class for displaying log messages in a GTK status bar."""

    def __init__(self, status=None, context=0):
        self.status = status
        self.context = context
        logging.Handler.__init__(self)

    def pull_status(self, msgid):
        """Remove specified msgid from the status stack."""
        self.status.remove_message(self.context, msgid)
        return False

    def push_status(self, msg, level):
        """Push the given msg onto the status stack, and defer removal."""
        delay = 3
        if level > 25:
            delay = 8
        msgid = self.status.push(self.context, msg)
        glib.timeout_add_seconds(delay, self.pull_status, msgid)
        return False

    def emit(self, record):
        """Emit log record to gtk main loop."""
        msg = self.format(record)
        glib.idle_add(self.push_status, msg, record.levelno)



"""Gemini (numeric) scoreboard sender.

 Output information to a pair of numeric, 9 character Gemini boards.
 Boards have the following layout:

   ---------------------
  | N N N   T:T:T:T:T:T |	N... is usually a rider no
   ---------------------	T... is time, punctuation may be ' ' , : or .

 Trackmeet events output the following formats:

 Bib, Rank & Time for flying 200, and elimination (show_brt):
 Content is the same on both boards

   ---------------------
  | 2 3     4   1 0.4 5 |	Rider 23, ranked 4th, 10.45 sec
   ---------------------

   ---------------------
  | 1 2 3               |	Rider 123 eliminated
   ---------------------

 "Dual" for ITT and pursuit race (show_dual):
 Allows for runtime up to 9:59.999, rolling time and downtime

   ---------------------
  | 1 2 3   3:4 5.6 7 8 |	Front straight, rider 123, 3:45.678
   ---------------------
  | 4 5     1:2 2.9 8 0 |	Back straight, rider 45, 1:22.980
   ---------------------

 "Running Time" for generic running time display (show_runtime):
 Content is the same on both boards, use bib to set first three chars

   ---------------------
  | 2 1     1 2:3 2.1   |	eg, 21 laps to go, elapsed 12:32.1
   ---------------------

 "Clock" for time of day (show_clock):
 Content is the same on both boards

   ---------------------
  |         1 2:4 5:3 7 |	Time of day: 12:45:37
   ---------------------

 "Lap" for display of lap on three-sided displays (show_lap)
 Content is the same on both boards

   ---------------------
  | 1 2 3   1 2:3 1:2 3 |	123 laps to go
   ---------------------

"""

import threading
import Queue
import logging
import serial
import sys

from metarace import unt4
from metarace import tod
from metarace import strops

# module logger
_log = logging.getLogger(u'metarace.gemini')
_log.setLevel(logging.DEBUG)

# dispatch thread queue commands
TCMDS = (u'EXIT', u'PORT', u'MSG')

# Character encoding
ENCODING = u'ascii'


class scbport(object):
    """Scoreboard communication port object."""

    def __init__(self, port=u'/dev/ttyUSB1'):
        self.__s = serial.Serial(port, 9600, rtscts=0, timeout=0.2)
        self.running = True

    def sendall(self, buf):
        """Send all of buf to port."""
        self.__s.write(buf.encode(ENCODING, u'ignore'))

    def close(self):
        """Shutdown socket object."""
        self.running = False
        try:
            self.__s.close()
        except:
            pass


GEMHEAD = unichr(unt4.SOH) + unichr(unt4.DC4)
GEMHOME = unichr(unt4.STX) + unichr(unt4.HOME)
GEMFOOT = unichr(unt4.EOT)


class gemini(threading.Thread):
    """Gemini sender thread."""

    def clear(self):
        """Clear scb."""
        self.bib = u''
        self.bib1 = u''
        self.rank = u''
        self.time = u''
        self.time1 = u''
        self.lap = u''
        self.lmsg = u''
        self.write(unt4.GENERAL_CLEARING.pack())

    def send_msg(self, msg, mtype=u'S', charoff=u'0', msg1=None):
        msg0 = msg
        if msg1 is None:
            msg1 = msg
        nmsg = (
            GEMHEAD  # front straight
            + mtype + u'0' + charoff + GEMHOME + msg0 + GEMFOOT +
            GEMHEAD  # back straight
            + mtype + u'0' + charoff + GEMHOME + unichr(unt4.LF)  # line 2
            + msg1 + GEMFOOT)
        if nmsg != self.lmsg:
            self.write(nmsg)
            self.lmsg = nmsg

    def reset_fields(self):
        """Clear out the gemini state fields."""
        self.bib = u''
        self.bib1 = u''
        self.rank = u''
        self.rank1 = u''
        self.time = u''
        self.time1 = u''

    def show_lap(self):
        self.lmsg = u''  # always write out lap to allow redraw
        lstr = strops.truncpad(self.lap, 3, u'r')
        msg = (lstr + unichr(unt4.STX) + lstr[0:2] + u':' + lstr[2] + lstr[0] +
               u':' + lstr[1:3] + u'.  ')
        self.send_msg(msg)

    def show_brt(self):
        msg = (
            strops.truncpad(self.bib, 3) + unichr(unt4.STX) +
            strops.truncpad(str(self.rank), 1) + u'  '  # the 'h:' padding
            + self.time.rjust(5))
        self.send_msg(msg)

    def show_dual(self):
        line0 = (strops.truncpad(self.bib, 3) + unichr(unt4.STX) +
                 strops.truncpad(self.time, 12, u'r', ellipsis=False))
        line1 = (strops.truncpad(self.bib1, 3) + unichr(unt4.STX) +
                 strops.truncpad(self.time1, 12, u'r', ellipsis=False))
        self.send_msg(line0, u'R', u'3', msg1=line1)  # xxx-5:02.6-x

    def set_rank(self, rank):
        if rank.isdigit() and len(rank) <= 1:
            self.rank = rank
        else:
            self.rank = u''

    def set_bib(self, bib, lane=False):  # True/1/something -> lane 1
        if lane:
            self.bib1 = bib
        else:
            self.bib = bib

    def set_time(self, time, lane=False):
        if lane:
            self.time1 = time
        else:
            self.time = time

    def show_clock(self):
        msg = (
            u'   '  # bib padding
            + unichr(unt4.STX) +
            strops.truncpad(self.time, 12, u'r', ellipsis=False))
        self.send_msg(msg, u'R')  # -2:34:56xxxx

    def rtick(self, ttod, places=3):
        """Convenience wrapper on set time/show runtime."""
        self.set_time(ttod.omstr(places))
        self.show_runtime()

    def dtick(self, ttod, places=3, lane=False):
        """Convenience wrapper on set time/show dualtime."""
        self.set_time(ttod.omstr(places), lane)
        self.show_dual()

    def ctick(self, ttod):
        """Convenience wrapper on set time/show clock."""
        self.set_time(ttod.omstr(0))
        self.show_clock()

    def set_lap(self, lap=u''):
        """Set and show the provided lap."""
        self.lap = unicode(lap)
        self.show_lap()

    def show_runtime(self):
        msg = (strops.truncpad(self.bib, 3) + unichr(unt4.STX) +
               strops.truncpad(self.time, 12, u'r', ellipsis=False))
        self.send_msg(msg, u'R', u'2')  # xxx-5:02.6-x

    def __init__(self, port=None):
        """Constructor."""
        threading.Thread.__init__(self)
        self.name = u'gemini'
        self.port = None
        self.ignore = False
        self.queue = Queue.Queue()
        self.running = False
        self.bib = u''
        self.bib1 = u''
        self.rank = u''
        self.rank1 = u''
        self.time = u''
        self.time1 = u''
        self.lap = u''
        self.lmsg = u''
        if port is not None:
            self.setport(port)

    def write(self, msg=None):
        """Send the provided msg to the scoreboard."""
        self.queue.put_nowait((u'MSG', msg))

    def exit(self, msg=None):
        """Request thread termination."""
        self.running = False
        self.queue.put_nowait((u'EXIT', msg))

    def wait(self):
        """Suspend calling thread until cqueue is empty."""
        self.queue.join()

    def setport(self, port=None):
        """Dump command queue content and (re)open port."""
        try:
            while True:
                self.queue.get_nowait()
                self.queue.task_done()
        except Queue.Empty:
            pass
        self.queue.put_nowait((u'PORT', port))

    def set_ignore(self, ignval=False):
        """Set or clear the ignore flag."""
        self.ignore = bool(ignval)

    def connected(self):
        """Return true if SCB connected."""
        return self.port is not None and self.port.running

    def run(self):
        """Called via threading.Thread.start()."""
        self.running = True
        _log.debug(u'Starting')
        while self.running:
            m = self.queue.get()
            self.queue.task_done()
            try:
                if m[0] == u'MSG' and not self.ignore and self.port:
                    #_log.debug(u'Send: %r', m[1])
                    self.port.sendall(m[1])
                elif m[0] == u'EXIT':
                    _log.debug(u'Request to close: %s', m[1])
                    self.running = False
                elif m[0] == u'PORT':
                    if self.port is not None:
                        self.port.close()
                        self.port = None
                    if m[1] is not None and m[1] != u'' and m[1] != u'NULL':
                        _log.debug(u'Re-Connect port: %s', m[1])
                        self.port = scbport(m[1])
                    else:
                        _log.debug(u'Not connected.')

            except IOError as e:
                _log.error(u'IO Error: %s', e)
                if self.port is not None:
                    self.port.close()
                self.port = None
            except Exception as e:
                _log.error(u'%s: %s', e.__class__.__name__, e)
        if self.port is not None:
            self.port.close()
        _log.debug(u'Exiting')



"""A collection of tools for preparing cycle race results."""

from __future__ import (division, absolute_import)

import os
import gtk
import glib
import gobject
import logging
import fcntl
import errno
from tempfile import NamedTemporaryFile
from shutil import copyfile
from metarace import jsonconfig

VERSION = u'1.12.3'
LIB_PATH = os.path.realpath(os.path.dirname(__file__))
UI_PATH = os.path.join(LIB_PATH, u'ui')
DB_PATH = os.path.join(LIB_PATH, u'data')
DATA_PATH = os.path.realpath(
    os.path.expanduser(os.path.join(u'~', u'Documents', u'metarace')))
DEFAULTS_NAME = u'default'
DEFAULTS_PATH = os.path.join(DATA_PATH, DEFAULTS_NAME)
LOGO_FILE = u'metarace_icon.svg'
SYSCONF_FILE = u'metarace.json'
PDF_TEMPLATE_FILE = u'pdf_template.json'
HTML_TEMPLATE_FILE = u'html_template.json'
LOGFILEFORMAT = u'%(asctime)s %(levelname)s:%(name)s: %(message)s'
LOGFORMAT = u'%(levelname)s:%(name)s: %(message)s'
LOGLEVEL = logging.DEBUG  # default console log level
sysconf = jsonconfig.config()  # system-defaults, populated by init() method
_log = logging.getLogger(u'metarace')
_log.setLevel(logging.DEBUG)


def init(withgtk=False):
    """Shared metarace program initialisation."""
    _log.debug(u'library init withgtk=%r', withgtk)
    copyconf = mk_data_path()

    # Set global logging options
    logging._srcfile = None
    logging.logThreads = 0
    logging.logProcesses = 0

    # read in system configuration
    conffile = default_file(SYSCONF_FILE)
    try:
        if os.path.exists(conffile):
            _log.debug(u'Loading system defaults from %r', conffile)
            with open(conffile, 'rb') as f:
                sysconf.read(f)
        else:
            _log.info(u'System default file not found %r', conffile)
    except Exception as e:
        _log.error(u'Error reading system config from %r: %s', conffile, e)

    # Do GTK init if required
    if withgtk:
        # Initialise threading in glib
        try:
            glib.threads_init()
        except Exception:
            _log.debug(u'glib thread init failed, using gobject: %s', e)
            gobject.threads_init()

        # Initialise threading in GDK
        gtk.gdk.threads_init()

        try:
            # fix the default gnome menubar accelerator mapping
            mset = gtk.settings_get_default()
            mset.set_string_property('gtk-menu-bar-accel', 'F24', 'override')
            gtk.window_set_default_icon_from_file(default_file(LOGO_FILE))
        except Exception as e:
            _log.debug(u'GTK init error: %s', e)
    # if required, create a new system default file
    if copyconf:
        _log.info(u'Creating default system config %s', SYSCONF_FILE)
        with savefile(os.path.join(DEFAULTS_PATH, SYSCONF_FILE)) as f:
            sysconf.write(f)


def mainloop():
    """Call into the gtk main loop."""
    gtk.gdk.threads_enter()
    gtk.main()
    gtk.gdk.threads_leave()


def mk_data_path():
    """Create a shared data path if it does not yet exist."""
    ret = False
    if not os.path.exists(DATA_PATH):
        _log.info(u'Creating data directory: %r', DATA_PATH)
        os.makedirs(DATA_PATH)
    if not os.path.exists(DEFAULTS_PATH):
        _log.info(u'Creating system defaults directory: %r', DEFAULTS_PATH)
        os.makedirs(DEFAULTS_PATH)
        ret = True  # flag copy of config back to defaults path
    return ret


def config_path(configpath=None):
    """Clean and check argument for a writeable meet configuration path."""
    ret = None
    if configpath is not None:
        # sanitise into expected config path
        ret = configpath
        if not os.path.isdir(ret):
            ret = os.path.dirname(ret)  # assume dangling path contains file
        ret = os.path.realpath(ret)
        _log.debug(u'Checking for meet %r using %r', configpath, ret)
        # then check if the path exists
        if not os.path.exists(ret):
            try:
                _log.info(u'Creating meet folder %r', ret)
                os.makedirs(ret)
            except Exception as e:
                _log.error(u'Unable to create folder %r: %s', ret, e)
                ret = None
        # check the path is writable
        if ret is not None:
            try:
                _log.debug(u'Checking folder %r for write access', ret)
                with NamedTemporaryFile(dir=ret, prefix=u'.chkwrite_') as f:
                    pass
            except Exception as e:
                _log.error(u'Unable to access meet folder %r: %s', ret, e)
                ret = None
    return ret


def default_file(filename=u''):
    """Return a path to the named file.

    Path components are stripped, then the the following locations
    are checked in order to find the first instance of filename:
        - current working directory
        - DEFAULTS_PATH
        - DB_PATH
    """
    filename = os.path.basename(filename)
    if filename in [u'..', u'.', u'', None]:
        return None

    ret = filename
    if os.path.exists(filename):
        pass
    else:
        check = os.path.join(DEFAULTS_PATH, filename)
        if os.path.exists(check):
            ret = check
        else:
            ## TODO: replace with lookup in package resources
            check = os.path.join(DB_PATH, filename)
            if os.path.exists(check):
                ret = check
    return ret


class savefile(object):
    """Tempfile-backed save file contextmanager."""

    def __init__(self, filename, tempdir=u'.'):
        self.__sfile = filename
        self.__path = tempdir
        self.__tfile = NamedTemporaryFile(mode='wb',
                                          suffix=u'.tmp',
                                          prefix=u'sav_',
                                          dir=self.__path,
                                          delete=False)

    def __enter__(self):
        return self.__tfile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__tfile.close()
        if exc_type is not None:
            return False  # raise exception
        # otherwise, file is saved ok in temp file
        os.chmod(self.__tfile.name, 0o644)
        try:
            os.rename(self.__tfile.name, self.__sfile)
            _log.debug(u'os.rename: %r,%r', self.__tfile.name, self.__sfile)
        except OSError as e:
            _log.debug(u'os.rename failed: %s', e)
            copyfile(self.__tfile.name, self.__sfile)
            _log.warn(u'Un-safely moved file: %r', self.__sfile)
            os.unlink(self.__tfile.name)
        return True


def lockpath(configpath):
    """Open an advisory lock file in the meet config path."""
    lf = None
    lfn = os.path.join(configpath, u'.lock')
    try:
        lf = open(lfn, 'a+b')
        fcntl.flock(lf, fcntl.LOCK_EX | fcntl.LOCK_NB)
        _log.debug(u'Config lock %r acquired', lfn)
    except Exception as e:
        if lf is not None:
            lf.close()
            lf = None
        _log.error(u'Unable to acquire config lock %r: %s', lfn, e)
    return lf


def unlockpath(configpath, lockfile):
    """Release advisory lock and remove lock file."""
    lfn = os.path.join(configpath, u'.lock')
    os.unlink(lfn)
    lockfile.close()
    _log.debug(u'Config lock %r released', lfn)
    return None


def about_dlg(window):
    """Display shared about dialog."""
    dlg = gtk.AboutDialog()
    dlg.set_transient_for(window)
    dlg.set_name(u'metarace')
    dlg.set_version(VERSION)
    dlg.set_copyright(
        u'Copyright \u00a9 2012-2022 Nathan Fraser and contributors')
    dlg.set_comments(u'Cycle Race Toolkit')
    dlg.set_license(LICENSETEXT)
    dlg.run()
    dlg.destroy()


LICENSETEXT = """
MIT License

Copyright (c) 2012-2023 Nathan Fraser and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



"""Rider name database helper object.

This module provides a simple object helper for manipulating the
rider namebank database. Methods are provided for retrieving rows
by license no and for searching by name.

namebank objects are intended to be used in a 'with' context eg:

  with namebank() as n:
     matches = n.search(u'first', u'last')

NOTE: namebank keys are strictly ASCII, all other fields are unicode.

"""

import os
import shelve
import datetime
import metarace
from metarace import ucsv

COL_ID = 0
COL_FIRST = 1
COL_LAST = 2
COL_CLUB = 3
COL_CAT = 4
COL_STATE = 5  ## origin
COL_DOB = 6
COL_GENDER = 7
COL_NATION = 8  # UCI Nationality as per 1.1.033
COL_EXPIRE = 9
COL_PARA = 10
COL_UCI = 11  # UCI ID as string

THISYEAR = datetime.date.today().year  # today returns local date


def get_ioc_codes():
    """Load and return a map of IOC codes->Countries."""
    iocmap = {}
    with open(metarace.default_file(u'IOC_Codes.csv'), 'rb') as f:
        cr = ucsv.UnicodeReader(f)
        for r in cr:
            if len(r) > 1 and u'CODE' not in r[0].upper():
                iocmap[r[0]] = r[1]
    return iocmap


def bgswitch(gender):
    ret = u'B'
    if gender.lower() == u'w':
        ret = u'G'
    return ret


def dob2lcat(gender, dob, reg=u'opn', para=None, year=None):
    """Return license category."""
    ty = THISYEAR
    if year is not None:
        ty = year
    lcat = u'U/C'
    if u'mas' in reg.lower():
        reg = u'mas'
    elif reg.lower() == u'n/c':
        return u'N/C'  # non-racing membership
    else:
        reg = u'opn'
    yob = dob.split(u'-')[0]
    if para:
        # para cycling is weird
        pcat = para
        if u'B' in para:
            pcat = u'B'
        lcat = gender + pcat
    elif yob.isdigit() and len(yob) == 4:
        age = ty - int(yob)
        if age < 13:
            gender = bgswitch(gender)
        if age < 11:
            lcat = u'J' + gender + u'11'
        elif age < 13:
            lcat = u'J' + gender + u'13'
        elif age < 15:
            lcat = u'J' + gender + u'15'
        elif age < 17:
            lcat = u'J' + gender + u'17'
        elif age < 19:
            lcat = u'J' + gender
        elif gender != u'W' and age < 23:
            lcat = gender + u'23'
        else:
            if reg == u'mas':  # Masters license
                mcat = u''
                # determine age cat
                if age < 35:
                    mcat = u'1'
                elif age < 40:
                    mcat = u'2'
                elif age < 45:
                    mcat = u'3'
                elif age < 50:
                    mcat = u'4'
                elif age < 55:
                    mcat = u'5'
                elif age < 60:
                    mcat = u'6'
                elif age < 65:
                    mcat = u'7'
                elif age < 70:
                    mcat = u'8'
                elif age < 75:
                    mcat = u'9'
                else:
                    mcat = u'10'
                lcat = gender + u'MAS' + mcat
            else:
                lcat = u'Elite' + gender
    else:
        lcat = None
    return lcat


def search_name(namestr):
    return namestr.strip().lower().encode(u'ascii', u'ignore')


class namebank(object):
    """Namebank storage and search module.

    The namebank object maintains a persistent storage of rider rows
    with the following structure:

     KEY - String: CA license 'no' or rider ID
     VAL - Array: [as per COL*]

    Searching by name uses an internal index to facilitate speedy
    return of matching riders.

    Internally two python shelve objects are used to map search keys
    to lists of rider info (for the namebank) or rider ids (for the
    name index).

    """

    def __init__(self):
        """Constructor."""
        self.__open = False
        self.__nb = None
        self.__ind = None

    def open(self):
        """(Re)Open the namebank database files."""
        self.close()
        self.__nb = shelve.open(
            os.path.join(metarace.DEFAULTS_PATH, u'namebank'))
        self.__ind = shelve.open(
            os.path.join(metarace.DEFAULTS_PATH, u'nameindx'))
        self.__open = True

    def close(self):
        """Close the namebank database files."""
        if self.__nb is not None:
            self.__nb.close()
            self.__nb = None
        if self.__ind is not None:
            self.__ind.close()
            self.__ind = None
        self.__open = False

    def raw_search(self, skey):
        """Return an exact match from the index."""
        ret = None
        if skey in self.__ind:
            ret = [a for a in self.__ind[skey]]
        return ret

    def ksearch(self, nkey=u''):
        """Return riders matching a NRS key search."""
        # NRS Key: flllyyyy[n*]
        # where:
        #        f = lowercase first initial
        #        lll = lowercase 3 ascii from surname not unicode/punc
        #        yyyy = year of birth
        #        n* = disambiguation digits, ignored in search
        key = unicode(search_name(nkey)).ljust(12)
        ret = set()
        fs = key[0]
        ls = key[1:4]
        yob = key[4:8]
        da = key[8:]
        cl = self.search(fs, ls)
        for ri in cl:
            dyo = self.__nb[ri][COL_DOB][0:4]
            if dyo == yob:
                ret.add(ri)

        return ret

    def search(self, first=u'', last=u'', dob=None):
        """Return a set of matching rider ids from the namebank."""

        # reformat search strings
        fs = search_name(first)
        ls = search_name(last)

        # Build candidate id set
        cset = set()
        if fs[0:4] in self.__ind:
            cset = cset.union(self.__ind[fs[0:4]])
        if ls[0:4] in self.__ind:
            cset = cset.union(self.__ind[ls[0:4]])
        # filter candidates further on full search string
        fset = set()
        if len(first) > 0:
            for r in cset:
                doc = self.__nb[r][COL_DOB]
                fn = self.__nb[r][COL_FIRST]
                if search_name(fn).find(fs) == 0:
                    fset.add(r)  # mark r in first name set
        else:
            fset = cset  # 'empty' first matches all
        lset = set()
        if len(last) > 0:
            for r in cset:
                doc = self.__nb[r][COL_DOB]
                ln = self.__nb[r][COL_LAST]
                if search_name(ln).find(ls) == 0:
                    lset.add(r)  # mark r in last name set
        else:
            lset = cset

        # return intersection of fset and lset
        return fset.intersection(lset)

    def __iter__(self):
        """Iterate over all namebank records."""
        for key in self.__nb:
            yield (self.__nb[key])

    def __len__(self):
        """Called to implement the built-in function len()."""
        return len(self.__nb)

    def __getitem__(self, key):
        """Called to implement evaluation of self[key]."""
        return self.__nb[key]

    def __contains__(self, key):
        """Called to implement membership test operators."""
        return key in self.__nb

    def __enter__(self):
        """Enter the runtime context related to this object."""
        if not self.__open:
            self.open()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        """Exit the runtime context related to this object."""
        self.close()



"""DHI scoreboard sender class.

This module provides a thread object which collects
and dispatches DHI messages intended for a Galactica or
Caprica scoreboard.

"""

import threading
import Queue
import logging
import socket
import metarace
import serial

from metarace import unt4
from metarace import strops

# Caprica encoding is UTF-8, Galactica is undefined - probably CP1252
ENCODING = 'utf-8'
LINELEN = 24
PAGELEN = 7
SERIALBAUD = 115200

# dispatch thread queue commands
TCMDS = (u'EXIT', u'PORT', u'MSG')

# module log object
_log = logging.getLogger(u'metarace.sender')
_log.setLevel(logging.DEBUG)


class serialport(object):
    """Scoreboard communication port object."""

    def __init__(self, addr, baudrate):
        """Constructor.

        Parameters:

          addr -- serial device filename
          baudrate -- serial line speed

        """
        _log.debug(u'Serial connection %s @ %d baud.', addr, baudrate)
        self.__s = serial.Serial(addr, baudrate, rtscts=False)
        self.send = self.__s.write
        self.running = True

    def sendall(self, buf):
        """Send all of buf to port."""
        msglen = len(buf)
        sent = 0
        while sent < msglen:
            out = self.send(buf[sent:])
            sent += out

    def close(self):
        """Shutdown socket object."""
        self.running = False
        try:
            self.__s.close()
        except:
            pass


class scbport(object):
    """Scoreboard communication port object."""

    def __init__(self, addr, protocol):
        """Constructor.

        Parameters:

          addr -- socket style 2-tuple (host, port)
          protocol -- one of socket.SOCK_STREAM or socket.SOCK_DGRAM

        """
        self.__s = socket.socket(socket.AF_INET, protocol)
        if protocol == socket.SOCK_STREAM:
            # set the TCP 'no delay' option
            self.__s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            _log.debug(u'Opening TCP socket %s', repr(addr))
        else:  # assume Datagram (UDP)
            # enable broadcast send
            self.__s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            _log.debug(u'Opening UDP socket %s', repr(addr))
        self.__s.connect(addr)
        self.send = self.__s.send
        self.running = True

    def sendall(self, buf):
        """Send all of buf to port."""
        msglen = len(buf)
        sent = 0
        #_log.debug(u'Sending: %r', buf)
        while sent < msglen:
            out = self.send(buf[sent:])
            if out == 0:
                raise socket.error(u'DHI sender socket broken')
            sent += out

    def close(self):
        """Shutdown socket object."""
        self.running = False
        try:
            self.__s.shutdown(socket.SHUT_RDWR)
        except:
            pass


def mkport(port=None):
    """Create a new scbport object.

    port is a string specifying the address as follows:

        [PROTOCOL:]ADDRESS[:PORT]

    Where:

        PROTOCOL :: TCP or UDP	(optional)
        ADDRESS :: hostname or IP address or device file
        PORT :: port name or number (optional)

    """
    nprot = socket.SOCK_STREAM  # default is TCP
    naddr = u'localhost'  # default is localhost
    nport = 1946  # default is 1946 (Caprica DHI)

    # import system defaults if required
    if metarace.sysconf.has_option(u'sender', u'portspec'):
        if not port or port == u'DEFAULT':
            port = metarace.sysconf.get(u'sender', u'portspec')

    if port == u'DEBUG':  # force use of the hardcoded UDP endpoint
        _log.debug(u'Using debug port: UDP:localhost:5060')
        nprot = socket.SOCK_DGRAM
        naddr = u'localhost'
        nport = 5060
    else:
        vels = [u'TCP', u'localhost', u'1946']
        aels = port.translate(strops.PRINT_UTRANS).strip().split(u':')
        if len(aels) == 3:
            vels[0] = aels[0].upper()
            vels[1] = aels[1]
            vels[2] = aels[2]
        elif len(aels) == 2:
            if aels[0].upper() in [u'TCP', u'UDP']:
                # assume PROT:ADDR
                vels[0] = aels[0].upper()
                vels[1] = aels[1]
            else:
                vels[1] = aels[0]
                vels[2] = aels[1]
        elif len(aels) == 1:
            vels[1] = aels[0]
        else:
            raise socket.error(u'Invalid port specification string')

        # 'import' the vels...
        if vels[0] == u'TCP':
            nprot = socket.SOCK_STREAM
            nport = 1946
        elif vels[0] == u'UDP':
            nprot = socket.SOCK_DGRAM
            nport = 5060
        else:
            raise socket.error(u'Invalid protocol specified.')
        naddr = vels[1]
        # override port if supplied
        if vels[2].isdigit():
            nport = int(vels[2])
        else:
            nport = socket.getservbyname(vels[2])

    # split port string into [PROTOCOL:]ADDR[:PORT]
    if u'/dev/' in naddr:
        # assume device file for a serial port
        baud = SERIALBAUD
        if metarace.sysconf.has_option(u'sender', u'serialbaud'):
            baud = strops.confopt_posint(
                metarace.sysconf.get(u'sender', u'serialbaud'), SERIALBAUD)
        return serialport(naddr, baud)
    else:
        return scbport((naddr, nport), nprot)


def mksender(port=None):
    cols = LINELEN
    rows = PAGELEN
    encoding = ENCODING
    if metarace.sysconf.has_option(u'sender', u'linelen'):
        cols = metarace.sysconf.get(u'sender', u'linelen')
    if metarace.sysconf.has_option(u'sender', u'pagelen'):
        rows = metarace.sysconf.get(u'sender', u'pagelen')
    if metarace.sysconf.has_option(u'sender', u'encoding'):
        encoding = metarace.sysconf.get(u'sender', u'encoding')
    ret = sender(port)
    ret.linelen = cols
    ret.pagelen = rows
    ret.encoding = encoding
    return ret


class sender(threading.Thread):
    """Caprica/Galactica DHI sender thread."""

    def clrall(self):
        """Clear all lines in DHI database."""
        self.sendmsg(unt4.GENERAL_CLEARING)

    def clrline(self, line):
        """Clear the specified line in DHI database."""
        self.sendmsg(unt4.unt4(xx=0, yy=int(line), erl=True))

    def setline(self, line, msg):
        """Set the specified DHI database line to msg."""
        msg = strops.truncpad(msg, self.linelen, u'l', False)
        self.sendmsg(unt4.unt4(xx=0, yy=int(line), erl=True, text=msg))

    def flush(self):
        """Send an empty update to force timeout clock to zero."""
        self.sendmsg(unt4.GENERAL_EMPTY)

    def linefill(self, line, char=u'_'):
        """Use char to fill the specified line."""
        msg = char * self.linelen
        self.sendmsg(unt4.unt4(xx=0, yy=int(line), text=msg))

    def postxt(self, line, oft, msg):
        """Position msg at oft on line in DHI database."""
        self.sendmsg(unt4.unt4(xx=int(oft), yy=int(line), text=msg))

    def setoverlay(self, newov):
        """Request overlay newov to be displayed on the scoreboard."""
        if self.curov != newov:
            self.sendmsg(newov)
            self.curov = newov

    def __init__(self, port=None):
        """Constructor."""
        threading.Thread.__init__(self)
        self.daemon = True
        self.name = u'sender'
        self.port = None
        self.linelen = LINELEN  # overridden by mksender
        self.pagelen = PAGELEN  # overridden by mksender
        self.encoding = ENCODING  # overridden by mksender
        self.ignore = False
        self.curov = None
        self.queue = Queue.Queue()
        self.running = False
        if port is not None:
            self.setport(port)

    def sendmsg(self, unt4msg=None):
        """Pack and send a unt4 message to the DHI."""
        self.queue.put_nowait((u'MSG', unt4msg.pack()))

    def write(self, msg=None):
        """Send the provided raw msg to the DHI."""
        self.queue.put_nowait((u'MSG', msg))

    def exit(self, msg=None):
        """Request thread termination."""
        self.running = False
        self.queue.put_nowait((u'EXIT', msg))

    def wait(self):
        """Suspend calling thread until cqueue is empty."""
        self.queue.join()

    def setport(self, port=None):
        """Dump command queue contents and (re)open DHI port.

        Specify hostname and port for TCP connection as follows:

            tcp:hostname:16372

        Or use DEBUG for a fallback UDP socket:

	    UDP:localhost:5060

        """
        try:
            while True:
                self.queue.get_nowait()
                self.queue.task_done()
        except Queue.Empty:
            pass
        self.queue.put_nowait((u'PORT', port))

    def set_ignore(self, ignval=False):
        """Set or clear the ignore flag."""
        self.ignore = bool(ignval)

    def connected(self):
        """Return true if SCB connected."""
        return self.port is not None and self.port.running

    def run(self):
        self.running = True
        _log.debug(u'Starting')
        while self.running:
            m = self.queue.get()
            self.queue.task_done()
            try:
                if m[0] == u'MSG' and not self.ignore and self.port:
                    #_log.debug(u'Sending message : ' + repr(m[1]))
                    self.port.sendall(m[1].encode(self.encoding, 'replace'))
                elif m[0] == u'EXIT':
                    _log.debug(u'Request to close: %s', m[1])
                    self.running = False
                elif m[0] == u'PORT':
                    if self.port is not None:
                        self.port.close()
                        self.port = None
                    if m[1] not in [None, u'', u'none', u'NULL']:
                        _log.debug(u'Re-Connect port: %s', m[1])
                        self.port = mkport(m[1])
                        self.curov = None
                    else:
                        _log.debug(u'Not connected.')

            except IOError as e:
                _log.error(u'IO Error: %s', e)
                if self.port is not None:
                    self.port.close()
                self.port = None
            except Exception as e:
                _log.error(u'%s: %s', e.__class__.__name__, e)
        if self.port is not None:
            self.port.close()
        _log.info(u'Exiting')



"""Time of Day types and functions

Time of Day (tod) records are used to compute net times,
and to communicate context of timing events. Each tod object
includes the following properties:

	timeval	decimal number of seconds, to 4 places
	index	optional serial number or marker (from timing device)
	chan	optional timing channel number or indicator
	refid	optional transponder id or rider identifier
	source	optional source id of timing device that generated the tod

Two specific types are provided:

	tod	Time of Day, strictly >= 0 and less than 24 hours,
		arithmetic is mod 24 hours.
	agg	Aggregate time, may be greater than 24 hours,
		negative values permitted.

Supported arithmetic operations:

		Y:	tod	agg	int	decimal
	tod - Y		yes	no	no	no	*
	tod + Y		yes	no	no	no	*
	agg - Y		yes	yes	yes	yes
	agg + Y		yes	yes	yes	yes

	* Result is mod 24 hours

"""
from __future__ import division

import decimal
import re
import logging
from datetime import datetime
from bisect import bisect_left as _bisect

# module log object
_log = logging.getLogger(u'metarace.tod')
_log.setLevel(logging.DEBUG)

# Formatting and truncation constants
_QUANT_5PLACES = decimal.Decimal(u'0.00001')
_QUANT_4PLACES = decimal.Decimal(u'0.0001')
_QUANT_3PLACES = decimal.Decimal(u'0.001')
_QUANT_2PLACES = decimal.Decimal(u'0.01')
_QUANT_1PLACE = decimal.Decimal(u'0.1')
_QUANT_0PLACES = decimal.Decimal(u'1')
_QUANT = [
    _QUANT_0PLACES, _QUANT_1PLACE, _QUANT_2PLACES, _QUANT_3PLACES,
    _QUANT_4PLACES, _QUANT_5PLACES
]
_QUANT_FW = [2, 4, 5, 6, 7, 8]
_QUANT_TWID = [8, 10, 11, 12, 13, 14]
_QUANT_PAD = [u'     ', u'   ', u'  ', u' ', u'', u'']
_QUANT_OPAD = [u'    ', u'  ', u' ', u'', u'', u'']
_MILL = decimal.Decimal(1000000)


def now(index=u'', chan=u'CLK', refid=u'', source=u'host'):
    """Return a tod set to the current local time."""
    return tod(_now2dec(), index, chan, refid, source)


def mkagg(timeval=u''):
    """Return agg for given timeval or None"""
    ret = None
    if timeval is not None and timeval != u'':
        try:
            ret = agg(timeval)
        except Exception as e:
            _log.debug(u'mkagg() %s: %s', e.__class__.__name__, e)
    return ret


def mktod(timeval=u''):
    """Return tod for given timeval or None"""
    ret = None
    if timeval is not None and timeval != u'':
        try:
            ret = tod(timeval)
        except Exception as e:
            _log.debug(u'mktod() %s: %s', e.__class__.__name__, e)
    return ret


def _now2dec():
    """Create a decimal timevalue for the current local time."""
    dv = datetime.now()
    ret = (dv.microsecond / _MILL).quantize(_QUANT_4PLACES)
    ret += 3600 * dv.hour + 60 * dv.minute + dv.second
    return ret


def _dec2hm(dectod=None):
    """Return truncated time string in hours and minutes"""
    strtod = None
    if dectod is not None:
        if dectod >= 3600:  # 'HH:MM'
            strtod = u'{0}:{1:02}'.format(
                int(dectod) // 3600, (int(dectod) % 3600) // 60)
        else:  # 'M'
            strtod = u'{0}'.format(int(dectod) // 60)
    return strtod


def _dec2str(dectod=None, places=4, zeros=False, hoursep=u'h', minsep=u':'):
    """Return formatted string for given tod decimal value

    Optional argument 'zeros' will use leading zero chars
    up to 24 hours. eg:

             '00h00:01.2345'   zeros=True
                    '1.2345'   zeros=False
    """
    strtod = None
    if dectod is not None:
        sign = u''
        # quantize first to preserve down rounding
        dv = dectod.quantize(_QUANT[places], rounding=decimal.ROUND_FLOOR)
        if dv < 0:
            dv = dv.copy_negate()
            sign = u'-'
        if zeros or dv >= 3600:  # '-HhMM:SS.dcmz'
            fmt = u'{0}{1}{2}{3:02}{4}{5:0{6}}'
            if zeros:  # '-00h00:0S.dcmz'
                fmt = u'{0}{1:02}{2}{3:02}{4}{5:0{6}}'
            strtod = fmt.format(sign,
                                int(dv) // 3600, hoursep,
                                (int(dv) % 3600) // 60, minsep, dv % 60,
                                _QUANT_FW[places])
        elif dv >= 60:  # '-M:SS.dcmz'
            strtod = u'{0}{1}{2}{3:0{4}}'.format(sign,
                                                 int(dv) // 60, minsep,
                                                 dv % 60, _QUANT_FW[places])
        else:  # '-S.dcmz'
            strtod = u'{0}{1}'.format(sign, dv)
    return strtod


def _str2dec(timestr=u''):
    """Return decimal for given string.

    Attempt to match against patterns:

    	-HhMM:SS.dcmz		Canonical
    	-H:MM:SS.dcmz		Omega
    	-H:MM'SS"dcmz		Chronelec
    	-H-MM-SS.dcmz		Keypad entry
    """
    dectod = None
    timestr = timestr.strip()  # assumes string
    if timestr == u'now':
        dectod = _now2dec()
    else:
        m = re.match(
            r'^(-?)(?:(?:(\d+)[h:-])?(\d{1,2})[:\'-])?(\d{1,2}(?:[\.\"]\d+)?)$',
            timestr,
            flags=re.UNICODE)
        if m is not None:
            dectod = decimal.Decimal(m.group(4).replace(u'"', u'.'))
            dectod += decimal.Decimal(m.group(3) or 0) * 60
            dectod += decimal.Decimal(m.group(2) or 0) * 3600
            if m.group(1):  # negative sign
                dectod = dectod.copy_negate()
        else:
            dectod = decimal.Decimal(timestr)
    return dectod


def _tv2dec(timeval):
    """Convert the provided value into a decimal timeval for tod/agg."""
    ret = 0
    if isinstance(timeval, decimal.Decimal):
        ret = timeval
    elif isinstance(timeval, basestring):
        ret = _str2dec(timeval)
    elif isinstance(timeval, tod):
        # Discard context on supplied tod and copy decimal obj
        ret = timeval.timeval
    elif isinstance(timeval, float):
        # Round off float to max tod precision
        ret = decimal.Decimal(u'{0:0.4f}'.format(timeval))
    else:
        ret = decimal.Decimal(timeval)
    return ret


class tod(object):
    """A class for working with short time intervals using time of day"""

    def __init__(self,
                 timeval=0,
                 index=u'',
                 chan=u'TOD',
                 refid=u'',
                 source=u'host'):
        self.index = index
        self.chan = chan
        self.refid = refid
        self.source = source
        self.timeval = _tv2dec(timeval)
        if self.timeval < 0 or self.timeval >= 86400:
            raise ValueError(u'Time of day value not in range [0, 86400)')

    def __str__(self):
        """Return a normalised tod string"""
        return str(self.__unicode__())

    def __unicode__(self):
        """Return a normalised tod string"""
        return u'{0: >5} {1: <3} {2} {3} {4}'.format(self.index, self.chan,
                                                     self.timestr(4),
                                                     self.refid, self.source)

    def __repr__(self):
        """Return object representation string"""
        return "{5}({0}, {1}, {2}, {3}, {4})".format(repr(self.timeval),
                                                     repr(self.index),
                                                     repr(self.chan),
                                                     repr(self.refid),
                                                     repr(self.source),
                                                     self.__class__.__name__)

    def truncate(self, places=4):
        """Return a new truncated time value"""
        return self.__class__(timeval=self.timeval.quantize(
            _QUANT[places], rounding=decimal.ROUND_FLOOR),
                              chan=u'TRUNC')

    def as_hours(self, places=0):
        """Return decimal value in hours, truncated to the desired places"""
        return (self.timeval / 3600).quantize(_QUANT[places],
                                              rounding=decimal.ROUND_FLOOR)

    def as_minutes(self, places=0):
        """Return decimal value in minutes, truncated to the desired places"""
        return (self.timeval / 60).quantize(_QUANT[places],
                                            rounding=decimal.ROUND_FLOOR)

    def as_seconds(self, places=0):
        """Return decimal value in seconds, truncated to the desired places"""
        return self.timeval.quantize(_QUANT[places],
                                     rounding=decimal.ROUND_FLOOR)

    def timestr(self, places=4, zeros=False, hoursep=u'h', minsep=u':'):
        """Return time string component of the tod, whitespace padded"""
        return u'{0: >{1}}{2}'.format(
            _dec2str(self.timeval, places, zeros, hoursep, minsep),
            _QUANT_TWID[places], _QUANT_PAD[places])

    def omstr(self, places=3, zeros=False, hoursep=u':', minsep=u':'):
        """Return a 12 digit omega style time string"""
        if places > 3:
            places = 3  # Hack to clamp to 12 dig
        return u'{0: >{1}}{2}'.format(
            _dec2str(self.timeval, places, zeros, hoursep, minsep),
            _QUANT_TWID[places], _QUANT_OPAD[places])

    def meridiem(self, mstr=None, secs=True):
        """Return a 12hour time of day string with meridiem"""
        ret = None
        med = u'\u2006am'
        # unwrap timeval into a single 24hr period
        tv = self.timeval
        if tv >= 86400:
            tv = tv % 86400
        elif tv < 0:
            tv = 86400 - (tv.copy_abs() % 86400)

        # determine meridiem and adjust for display
        if tv >= 43200:
            med = u'\u2006pm'
        if mstr is not None:
            med = mstr
        tv = tv % 43200
        if tv < 3600:  # 12am/12pm
            tv += 43200
        if secs:
            ret = _dec2str(tv, 0, hoursep=u':', minsep=u':') + med
        else:
            ret = _dec2hm(tv) + med
        return ret

    def rawtime(self, places=4, zeros=False, hoursep=u'h', minsep=u':'):
        """Return time string of tod as string, without padding"""
        return _dec2str(self.timeval, places, zeros, hoursep, minsep)

    def speedstr(self, dist=200):
        """Return average speed estimate string for the provided distance"""
        if self.timeval == 0:
            return u'---.- km/h'
        return u'{0:5.1f} km/h'.format(3.6 * float(dist) / float(self.timeval))

    def rawspeed(self, dist=200):
        """Return an average speed estimate string without unit"""
        if self.timeval == 0:
            return u'-.-'
        return u'{0:0.1f}'.format(3.6 * float(dist) / float(self.timeval))

    def __lt__(self, other):
        if isinstance(other, tod):
            return self.timeval < other.timeval
        else:
            return self.timeval < other

    def __le__(self, other):
        if isinstance(other, tod):
            return self.timeval <= other.timeval
        else:
            return self.timeval <= other

    def __eq__(self, other):
        if isinstance(other, tod):
            return self.timeval == other.timeval
        else:
            return self.timeval == other

    def __ne__(self, other):
        if isinstance(other, tod):
            return self.timeval != other.timeval
        else:
            return self.timeval != other

    def __gt__(self, other):
        if isinstance(other, tod):
            return self.timeval > other.timeval
        else:
            return self.timeval > other

    def __ge__(self, other):
        if isinstance(other, tod):
            return self.timeval >= other.timeval
        else:
            return self.timeval >= other

    def __sub__(self, other):
        """Compute time of day subtraction and return a NET tod object"""
        if type(other) is not tod:  # Subclass must override this method
            return NotImplemented
        if self.timeval >= other.timeval:
            oft = self.timeval - other.timeval
        else:
            oft = 86400 - other.timeval + self.timeval
        return tod(timeval=oft, chan=u'NET')

    def __add__(self, other):
        """Compute time of day addition and return a new tod object"""
        if type(other) is not tod:  # Subclass must override this method
            return NotImplemented
        return tod(timeval=(self.timeval + other.timeval) % 86400, chan=u'SUM')

    def __pos__(self):
        """Unary + operation"""
        return self.__class__(self.timeval, chan=u'POS')

    def __abs__(self):
        """Unary absolute value"""
        return self.__class__(self.timeval.copy_abs(), chan=u'ABS')


class agg(tod):
    """Aggregate time type"""

    def __init__(self,
                 timeval=0,
                 index=u'',
                 chan=u'AGG',
                 refid=u'',
                 source=u'host'):
        self.index = index
        self.chan = chan
        self.refid = refid
        self.source = source
        self.timeval = _tv2dec(timeval)

    def __add__(self, other):
        """Compute addition and return aggregate"""
        if isinstance(other, tod):
            return agg(timeval=self.timeval + other.timeval, chan=u'AGG')
        elif isinstance(other, (int, decimal.Decimal)):
            return agg(timeval=self.timeval + other, chan=u'AGG')
        else:
            return NotImplemented

    def __sub__(self, other):
        """Compute subtraction and return aggregate"""
        if isinstance(other, tod):
            return agg(timeval=self.timeval - other.timeval, chan=u'AGG')
        elif isinstance(other, (int, decimal.Decimal)):
            return agg(timeval=self.timeval - other, chan=u'AGG')
        else:
            return NotImplemented

    def __neg__(self):
        """Unary - operation"""
        return self.__class__(self.timeval.copy_negate(), chan=u'AGG')


# TOD constants
ZERO = tod()
ONE = tod(u'1.0')
MINUTE = tod(u'1:00')
MAX = tod(u'23h59:59.9999')  # largest val possible for tod
MAXELAP = tod(u'23h30:00')  # max displayed elapsed time

# Fake times for special cases
FAKETIMES = {
    u'catch': tod(ZERO, chan=u'catch'),
    u'w/o': tod(ZERO, chan=u'w/o'),
    u'max': tod(MAX, chan=u'max'),
    u'ntr': tod(MAX, chan=u'ntr'),
    u'caught': tod(MAX, chan=u'caught'),
    u'rel': tod(MAX, chan=u'rel'),
    u'abort': tod(MAX, chan=u'abort'),
    u'otl': tod(MAX, chan=u'otl'),
    u'dnf': tod(MAX, chan=u'dnf'),
    u'dns': tod(MAX, chan=u'dns'),
    u'dsq': tod(MAX, chan=u'dsq'),
}
_extra = decimal.Decimal(u'0.00001')
_cof = decimal.Decimal(u'0.00001')
for _c in [
        u'ntr', u'caught', u'rel', u'abort', u'otl', u'dnf', u'dns', u'dsq'
]:
    FAKETIMES[_c].timeval += _cof
    _cof += _extra


class todlist(object):
    """ToD list helper class for managing splits and ranks"""

    def __init__(self, lbl=u''):
        self.__label = lbl
        self.__store = []

    def __iter__(self):
        return self.__store.__iter__()

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        return self.__store[key]

    def rank(self, bib, series=u''):
        """Return current 0-based rank for given bib"""
        ret = None
        count = 0
        i = 0
        lpri = None
        lsec = None
        for lt in self.__store:
            # scan times for updating ranks
            if lpri is not None:
                if lt[0] != lpri or lt[1] != lsec:
                    i = count
            # if rider matches, break
            if lt[0].refid == bib and lt[0].index == series:
                ret = i
                break
            lpri = lt[0]
            lsec = lt[1]
            count += 1
        return ret

    def clear(self):
        """Clear list"""
        self.__store = []
        return 0

    def remove(self, bib, series=u''):
        """Remove all times matching the supplied bib and optional series"""
        i = 0
        while i < len(self.__store):
            if (self.__store[i][0].refid == bib
                    and self.__store[i][0].index == series):
                del self.__store[i]
            else:
                i += 1
        return i

    def insert(self, pri=None, sec=None, bib=None, series=u'', prec=3):
        """Insert primary and secondary tod labeled with bib and series"""
        ret = None
        trunc = True
        if pri in FAKETIMES:  # re-assign a coded 'finish'
            pri = FAKETIMES[pri]
            trunc = False  # retain precision for primary comparison

        if isinstance(pri, tod):
            if bib is None:
                bib = pri.index
            if sec is None:
                sec = ZERO
            if trunc:
                pri = pri.truncate(prec)
                sec = sec.truncate(prec)
            rt0 = tod(pri, chan=self.__label, refid=bib, index=series)
            rt1 = tod(sec, chan=self.__label, refid=bib, index=series)
            ret = _bisect(self.__store, (rt0, rt1))
            self.__store.insert(ret, (rt0, rt1))
        return ret



"""Scoreboard 'windows'.

This module provides a number of animated plaintext scoreboard window
objects for the display of lists, times and transitions.

A scoreboard window is a stateful information block that
may or may not be animated. All types of windows have the following
interface:

 reset()	reset state to start (calls redraw)
 pause()	toggle paused state, returns next val
 redraw()	redraw fixed screen elements
 update()	advance animation by one 'frame', caller is
		expected to repeatedly call update at ~20Hz

Shared properties for all scbwins:

	scb	A sender thread handle

Per-class init func should not draw onto screen, only redraw()
or first call to update() will emit anything to scb surface.

"""

import time  # for time.strftime() :(
import logging

from metarace import strops
from metarace import unt4
from metarace import tod

_log = logging.getLogger(u'metarace.scbwin')
_log.setLevel(logging.DEBUG)

PAGE_INIT = 10  # delay before table data starts displaying
PAGE_DELAY = 60  # def tenths of sec to hold each page of table
PAGE_ROWOFT = 2  # first DHI row for table data
DATE_FMT = u'%a %d/%m/%y'


def get_dateline(width=32):
    dpart = time.strftime(DATE_FMT)
    tpart = tod.now().meridiem()
    ret = tpart
    totlen = len(tpart) + len(dpart)
    if totlen >= width:  # with a space
        ret = tpart.center(width)  # fall back to time
    else:
        ret = dpart + u' ' * (width - totlen) + tpart

    return ret


class scbwin(object):
    """Base class for all scoreboard windows.

    Classes inheriting from scbwin are required to override the
    update() and redraw() methods.

    """

    def __init__(self, scb=None):
        """Base class constructor."""
        self.paused = False
        self.scb = scb
        self.count = 0

    def reset(self):
        """Reset scbwin to initial state."""
        self.count = 0
        self.redraw()
        self.paused = False

    def pause(self, set=None):
        """Update the pause property.

        If param 'set' is not provided toggle the current pause state,
        otherwise update pause to equal 'set'.

        """
        if set is not None:
            self.paused = bool(set)
        else:
            self.paused = not self.paused
        return self.paused

    def redraw(self):
        """Virtual redraw method."""
        pass

    def update(self):
        """Virtual update method."""
        self.count += 1


class scbclock(scbwin):
    """Event clock window.

    Display event lines under a date and time string. Eg:

      012345678901234567890123
      Sat 15/02/34___2:12:12pm		'__' expands with W
      ------------------------
           CENTERED LINE 1
           CENTERED LINE 2
           CENTERED LINE 3

    """

    def __init__(self, scb=None, line1=u'', line2=u'', line3=u'', locstr=u''):
        scbwin.__init__(self, scb)
        self.line1 = line1
        self.line2 = line2
        self.line3 = line3
        self.locstr = locstr
        self.bodyoft = 1  # body text offset
        if self.scb.pagelen > 4:
            self.bodyoft = 2
        self.header = get_dateline(self.scb.linelen)

    def redraw(self):
        self.scb.setline(0, self.header)
        for i in range(1, self.scb.pagelen):
            self.scb.clrline(i)
        self.scb.setoverlay(unt4.OVERLAY_MATRIX)

    def update(self):
        """Animate the clock window.

        After an initial pause, animate the title lines onto
        the scorebord with approx 0.1s delay between lines.

        Date and time in the header are autmomatically updated
        from the system time.

        """
        if not self.paused:
            if self.count == 14:
                self.scb.setline(self.bodyoft,
                                 self.line1.strip().center(self.scb.linelen))
            if self.count == 16:
                self.scb.setline(self.bodyoft + 1,
                                 self.line2.strip().center(self.scb.linelen))
            if self.count == 18:
                self.scb.setline(self.bodyoft + 2,
                                 self.line3.strip().center(self.scb.linelen))
            if self.count == 20:
                self.scb.setline(self.bodyoft + 4,
                                 self.locstr.strip().center(self.scb.linelen))
            if self.count % 2 == 0:
                next = get_dateline(self.scb.linelen)
                if next != self.header:
                    self.scb.setline(0, next)
                    self.header = next
            self.count += 1


class scbtt(scbwin):
    """Pursuit/ITT/Teams Timer window.

    Display a pursuit/time trial timer window with two names
    and two time values. Time values are copied onto the overlay
    within the update() method. No time calculations are conducted,
    this class only displays the strings provided.

    Example:

        012345678901234567890123
              Prefix Info
        ------------------------
        12 Blackburn Team 1
	   A. RIDER___B.RIDER___*	[*suppressed on small display]
	   C. RIDER___D.RIDER___*
        >>>>>>>>(1) hh:mm:ss.dcm 
        10 Blackburn Team 2
	   A. RIDER___B.RIDER___*
	   C. RIDER___D.RIDER___*
        >>>>>>>>(3) hh:mm:ss.dcm 

    """

    def __init__(self,
                 scb=None,
                 header=u'',
                 line1=u'',
                 line2=u'',
                 subheader=u''):
        scbwin.__init__(self, scb)
        self.header = header.strip().center(self.scb.linelen)
        self.subheader = subheader.strip().center(self.scb.linelen)
        self.line1 = line1
        self.nextline1 = line1
        self.line2 = line2
        self.nextline2 = line2
        self.curt1 = u''
        self.nextt1 = u''
        self.curr1 = u''
        self.nextr1 = u''
        self.curt2 = u''
        self.nextt2 = u''
        self.curr2 = u''
        self.nextr2 = u''
        self.singleoft = 0

    def set_single(self):
        self.singleoft = 1

    def redraw(self):
        clrset = list(range(0, self.scb.pagelen))
        self.scb.setline(0, self.header)
        clrset.remove(0)
        self.scb.setline(1, self.subheader)
        clrset.remove(1)
        self.scb.setline(2 + self.singleoft, self.line1)
        clrset.remove(2 + self.singleoft)
        self.scb.setline(4 + self.singleoft, self.line2)
        clrset.remove(4 + self.singleoft)
        # only clear rows not already set above.
        for i in clrset:
            self.scb.clrline(i)
        self.nextline1 = self.line1
        self.nextline2 = self.line2
        self.curt1 = u''
        self.nextt1 = u''
        self.curr1 = u''
        self.nextr1 = u''
        self.curt2 = u''
        self.nextt2 = u''
        self.curr2 = u''
        self.nextr2 = u''
        self.scb.setoverlay(unt4.OVERLAY_MATRIX)

    def setline1(self, line1str=u''):
        """Replace the line 1 text."""
        self.nextline1 = line1str

    def setline2(self, line2str=u''):
        """Replace the line 2 text."""
        self.nextline2 = line2str

    def sett1(self, timestr=u''):
        """Set the next front straight time string."""
        self.nextt1 = timestr

    def sett2(self, timestr=u''):
        """Set the next back straight time string."""
        self.nextt2 = timestr

    def setr1(self, rank=u''):
        """Set the next front straight rank string."""
        self.nextr1 = rank

    def setr2(self, rank=u''):
        """Set the next back straight rank string."""
        self.nextr2 = rank

    def update(self):
        """If any time or ranks change, copy new value onto overlay."""
        if not self.paused:
            if self.curr1 != self.nextr1 or self.curt1 != self.nextt1:
                self.scb.setline(
                    3 + self.singleoft,
                    strops.truncpad(self.nextr1, self.scb.linelen - 13, 'r') +
                    u' ' + self.nextt1)
                self.curr1 = self.nextr1
                self.curt1 = self.nextt1
            if self.curr2 != self.nextr2 or self.curt2 != self.nextt2:
                self.scb.setline(
                    5,
                    strops.truncpad(self.nextr2, self.scb.linelen - 13, 'r') +
                    u' ' + self.nextt2)
                self.curr2 = self.nextr2
                self.curt2 = self.nextt2
            if self.line1 != self.nextline1:
                self.line1 = self.nextline1
                self.scb.setline(2 + self.singleoft, self.nextline1)
            if self.line2 != self.nextline2:
                self.line2 = self.nextline2
                self.scb.setline(4 + self.singleoft, self.nextline2)
            self.count += 1


class scbtimer(scbwin):
    """Sprint timer window with avg speed.

    Copy provided time strings into pre-determined fields
    on the overlay. No time calcs are performed - this module
    only works on strings.

    Example:

        012345678901234567890123
          Blahface Point Score
          intermediate sprint
        ------------------------
              200m: hh:mm:ss.000
               Avg:  xx.yyy km/h

    """

    def __init__(self,
                 scb=None,
                 line1=u'',
                 line2=u'',
                 timepfx=u'',
                 avgpfx=u'Avg:'):
        scbwin.__init__(self, scb)
        self.line1 = line1
        self.line2 = line2
        self.timepfx = timepfx
        self.avgpfx = avgpfx
        self.curtime = u''
        self.nexttime = u''
        self.curavg = u''
        self.nextavg = u''

    def redraw(self):
        clrset = list(range(0, self.scb.pagelen))
        self.scb.setline(0, self.line1.strip().center(self.scb.linelen))
        clrset.remove(0)
        self.scb.setline(1, self.line2.strip().center(self.scb.linelen))
        clrset.remove(1)
        #self.scb.setline(3, strops.truncpad(self.timepfx,
        #self.scb.linelen - 13, 'r'))
        #clrset.remove(3)
        for i in clrset:
            self.scb.clrline(i)

        self.curtime = u''
        self.nexttime = u''
        self.curavg = u''
        self.nextavg = u''
        self.scb.setoverlay(unt4.OVERLAY_MATRIX)

    def settime(self, timestr=u''):
        """Set the next time speed string."""
        self.nexttime = timestr

    def setavg(self, avgstr=u''):
        """Set the next average speed string."""
        self.nextavg = avgstr

    def update(self):
        """If time or avg change, copy new value onto overlay."""
        if not self.paused:
            if self.curtime != self.nexttime:
                #self.scb.postxt(3, self.scb.linelen - 13,
                #strops.truncpad(self.nexttime,12,'r'))
                self.scb.setline(
                    3,
                    strops.truncpad(self.timepfx, self.scb.linelen - 13, 'r') +
                    strops.truncpad(self.nexttime, 13, 'r'))
                self.curtime = self.nexttime
            if self.curavg != self.nextavg:
                self.scb.setline(
                    4,
                    strops.truncpad(self.avgpfx, self.scb.linelen - 13, 'r') +
                    strops.truncpad(self.nextavg, 12, 'r'))
                self.curavg = self.nextavg
            self.count += 1


class scbtest(scbwin):
    """A test pattern to check character and line sizes."""

    def redraw(self):
        hdrv = []
        for i in range(0, self.scb.linelen):
            hdrv.append(u'{0:x}'.format(i % 16))
        self.scb.setline(0, u''.join(hdrv))
        hdrv[1] = u':'
        hdrv[2] = u' '
        hdrv[3] = u' '
        for i in range(1, self.scb.pagelen):
            hdrv[0] = u'{0:x}'.format(i % 16)
            self.scb.setline(i, u''.join(hdrv))
        self.scb.setoverlay(unt4.OVERLAY_MATRIX)

    def update(self):
        if not self.paused:
            if self.count % 400 == 0:
                self.scb.flush()
            self.count += 1

    def __init__(self, scb=None):
        scbwin.__init__(self, scb)


class scbintsprint(scbwin):
    """Intermediate sprint window - scrolling table, with 2 headers.

    Parameters coldesc and rows as per scbtable)

    """

    def loadrows(self, coldesc=None, rows=None):
        self.rows = []
        if coldesc is not None and rows is not None:
            for row in rows:
                nr = u''
                oft = 0
                for col in coldesc:
                    if type(col) in [unicode, str]:
                        nr += col
                    else:
                        if len(row) > oft:  # space pad 'short' rows
                            nr += strops.truncpad(row[oft], col[0], col[1])
                        else:
                            nr += u' ' * col[0]
                        oft += 1
                self.rows.append(nr[0:32])
        self.nrpages = len(self.rows) // self.pagesz + 1
        if self.nrpages > 1 and len(self.rows) % self.pagesz == 0:
            self.nrpages -= 1
        # avoid hanging residual by scooting 2nd last entry onto
        # last page with a 'dummy' row, or scoot single line down by one
        if len(self.rows) % self.pagesz == 1:
            self.rows.insert(len(self.rows) - 2, u' ')

    def redraw(self):
        self.scb.setline(0, self.line1.strip().center(self.scb.linelen))
        self.scb.setline(1, self.line2.strip().center(self.scb.linelen))
        for i in range(2, self.scb.pagelen):
            self.scb.clrline(i)
        self.scb.setoverlay(unt4.OVERLAY_MATRIX)

    def update(self):
        if self.count % 2 == 0 and self.count > PAGE_INIT:  # wait ~1/2 sec
            lclk = (self.count - PAGE_INIT) // 2
            cpage = (lclk // self.delay) % self.nrpages
            pclk = lclk % self.delay
            if pclk < self.pagesz + 1:
                if pclk != self.pagesz:
                    self.scb.clrline(self.rowoft + pclk)
                elif self.nrpages == 1:
                    self.count += 1
                    self.paused = True  # no animate on single page
                if pclk != 0:
                    roft = self.pagesz * cpage + pclk - 1
                    if roft < len(self.rows):
                        self.scb.setline(self.rowoft + pclk - 1,
                                         self.rows[roft])
        if not self.paused:
            self.count += 1

    def __init__(self,
                 scb=None,
                 line1=u'',
                 line2=u'',
                 coldesc=None,
                 rows=None,
                 delay=PAGE_DELAY):
        scbwin.__init__(self, scb)
        self.pagesz = 4
        self.nrpages = 0
        self.delay = delay
        self.rowoft = 3  # check for < 7

        # prepare header
        self.line1 = line1[0:self.scb.linelen]
        self.line2 = line2[0:self.scb.linelen]

        # load rows
        self.rows = []  # formatted rows
        self.loadrows(coldesc, rows)


class scbtable(scbwin):
    """A self-looping info table.

    Displays 'pages' of rows formatted to coldesc:
   
    Coldesc: set of column tuples, each containing a field width
             as integer and the string 'l' or 'r' for left
             or right space padded, or a string constant
   
	       [(fieldwidth, l|r)|'str' ...]
   
    Example:  [(3,'r'), ' ', '(20,'l')]
 		   Three columns:
			   1: 3 character str padded to right
			   2: constant string ' '
			   3: 20 character str padded to left

    ADDED: timepfx and timestr for appending a time field to results

    """

    def loadrows(self, coldesc=None, rows=None):
        self.rows = []
        if coldesc is not None and rows is not None:
            for row in rows:
                nr = u''
                oft = 0
                for col in coldesc:
                    if type(col) in [unicode, str]:  # HACK py<3.0
                        nr += col
                    else:
                        if len(row) > oft:  # space pad 'short' rows
                            nr += strops.truncpad(row[oft], col[0], col[1])
                        else:
                            nr += u' ' * col[0]
                        oft += 1
                self.rows.append(nr)  # truncation in sender ok
        self.nrpages = len(self.rows) // self.pagesz + 1
        if self.nrpages > 1 and len(self.rows) % self.pagesz == 0:
            self.nrpages -= 1
        # avoid hanging residual by scooting 2nd last entry onto
        # last page with a 'dummy' row, or scoot single line down by one
        if len(self.rows) % self.pagesz == 1:
            self.rows.insert(len(self.rows) - 2, u' ')

    def redraw(self):
        self.scb.setline(0, self.header.center(self.scb.linelen))
        j = 1
        if self.rowoft == 2:  # space for subheader
            self.scb.setline(1, self.subhead.center(self.scb.linelen))
            j = 2
        for i in range(j, self.scb.pagelen):
            self.scb.clrline(i)
        self.scb.setoverlay(unt4.OVERLAY_MATRIX)

    def update(self):
        # if time field set and not a round number of rows, append
        # time line to last row of last page
        if self.count % 2 == 0 and self.count > PAGE_INIT:  # wait ~1/2 sec
            lclk = (self.count - PAGE_INIT) // 2
            cpage = (lclk // self.delay) % self.nrpages
            pclk = lclk % self.delay
            # special case for single page results to hold page w/ Caprica
            if self.nrpages == 1 and lclk >= self.delay:
                if pclk == 0:
                    self.scb.flush()
            elif pclk < self.pagesz + 1:
                if pclk != self.pagesz:
                    self.scb.clrline(self.rowoft + pclk)
                else:  # end of page
                    #if self.nrpages == 1:
                    #self.count += 1
                    #self.paused = True # no further animate on single page
                    if self.timestr is not None:
                        self.scb.setline(
                            self.rowoft + pclk,
                            strops.truncpad(self.timepfx,
                                            self.scb.linelen - 13, 'r') +
                            u' ' + self.timestr[0:12])
                if pclk != 0:
                    roft = self.pagesz * cpage + pclk - 1
                    if roft < len(self.rows):
                        self.scb.setline(self.rowoft + pclk - 1,
                                         self.rows[roft])
        if not self.paused:
            self.count += 1

    def __init__(self,
                 scb=None,
                 head=u'',
                 subhead=u'',
                 coldesc=None,
                 rows=None,
                 pagesz=None,
                 timepfx=u'',
                 timestr=None,
                 delay=PAGE_DELAY):
        scbwin.__init__(self, scb)
        # set page size
        self.pagesz = self.scb.pagelen - 2
        # page row offset ... hardcoded for now
        self.rowoft = 2
        self.nrpages = 0
        self.delay = delay
        self.timestr = timestr
        self.timepfx = timepfx
        if pagesz and pagesz > 5:
            self.pagesz = 6  # grab a line from the top
            self.rowoft = 2

        if self.timestr is not None:
            self.pagesz -= 1  # grab a line from the bottom

        # prepare header -> must be preformatted
        if pagesz < 6:  # this is a hack for the madison, maybe replace?
            self.header = head[0:self.scb.linelen].strip()
        else:
            self.header = head[0:self.scb.linelen]
        self.subhead = subhead

        # load rows
        self.rows = []  # formatted rows
        self.loadrows(coldesc, rows)
